import inspect
from importlib.metadata import entry_points


def test_ep_override_methods():
    epg = entry_points(group="fromager.override_methods")
    assert epg
    for name in epg.names:
        func = epg[name].load()
        assert inspect.isfunction(func)



import logging
import pathlib

import click
import rich
import rich.box
from packaging.requirements import Requirement
from rich.table import Table

from fromager import (
    context,
    requirements_file,
)
from fromager.dependency_graph import DependencyGraph

logger = logging.getLogger(__name__)


@click.command()
@click.argument("req_file", type=click.Path(exists=True, path_type=pathlib.Path))
@click.argument("graph_file", type=click.Path(exists=True, path_type=pathlib.Path))
@click.pass_obj
def stats(
    wkctx: context.WorkContext,
    req_file: pathlib.Path,
    graph_file: pathlib.Path,
) -> None:
    """Show statistics about packages in a build

    REQ_FILE is the requirements.txt file used for the build

    GRAPH_FILE is the fromager graph.json file produced by the build

    Shows a table with various statistics about the packages including
    counts of unique packages, constraints, configurations, prebuilt packages,
    patches, and build plugins.
    """

    # Read requirements.txt file
    requirements_count = 0
    try:
        requirement_lines = list(requirements_file.parse_requirements_file(req_file))
        requirements_count = len(requirement_lines)
    except Exception as e:
        logger.error(f"Failed to read requirements file {req_file}: {e}")
        return

    # Read graph.json file
    try:
        graph = DependencyGraph.from_file(graph_file)
    except Exception as e:
        logger.error(f"Failed to read graph file {graph_file}: {e}")
        return

    # Extract unique package names from graph
    unique_packages: set[str] = set()
    prebuilt_packages: set[str] = set()

    for node in graph.get_all_nodes():
        # Skip the root node which has empty name
        if node.canonicalized_name:
            unique_packages.add(node.canonicalized_name)
            if node.pre_built:
                prebuilt_packages.add(node.canonicalized_name)

    # Analyze package configurations
    constrained_packages: set[str] = set()
    configured_packages: set[str] = set()
    patched_packages: set[str] = set()
    plugin_packages: set[str] = set()

    for package_name in unique_packages:
        try:
            req = Requirement(package_name)
            pbi = wkctx.package_build_info(req)

            # Check if package has constraints
            try:
                constraint = wkctx.constraints.get_constraint(package_name)
                if constraint is not None:
                    constrained_packages.add(package_name)
            except Exception:
                pass

            # Check if package has any fromager configuration
            try:
                # Use the has_config attribute from PackageSettings which indicates
                # whether the package has a configuration file
                ps = wkctx.settings.package_setting(package_name)
                if ps.has_config:
                    configured_packages.add(package_name)
            except Exception:
                pass

            # Check if package has patches
            try:
                all_patches = pbi.get_all_patches()
                if all_patches:  # Check if the patch map is not empty
                    patched_packages.add(package_name)
            except Exception:
                pass

            # Check if package has build plugins
            try:
                if pbi.plugin is not None:
                    plugin_packages.add(package_name)
            except Exception:
                pass

        except Exception as e:
            logger.debug(f"Could not analyze package {package_name}: {e}")
            continue

    # Create stats table
    table = Table(title="Build Statistics", box=rich.box.MARKDOWN, title_justify="left")
    table.add_column("Metric", justify="left", no_wrap=True)
    table.add_column("Count", justify="right", no_wrap=True)
    table.add_column("Percentage", justify="right", no_wrap=True)

    total_packages = len(unique_packages)

    def format_percentage(count: int, total: int) -> str:
        if total == 0:
            return "0.0%"
        return f"{(count / total) * 100:.1f}%"

    table.add_row("Requirements in requirements.txt", str(requirements_count), "-")
    table.add_row("Constraints defined", str(len(list(wkctx.constraints))), "-")
    table.add_row("Unique packages in build", str(total_packages), "100.0%")
    table.add_row(
        "Packages in the build with constraints",
        str(len(constrained_packages)),
        format_percentage(len(constrained_packages), total_packages),
    )
    table.add_row(
        "Pre-built packages",
        str(len(prebuilt_packages)),
        format_percentage(len(prebuilt_packages), total_packages),
    )
    table.add_row(
        "Packages with any fromager config",
        str(len(configured_packages)),
        format_percentage(len(configured_packages), total_packages),
    )
    table.add_row(
        "Packages with patches",
        str(len(patched_packages)),
        format_percentage(len(patched_packages), total_packages),
    )
    table.add_row(
        "Packages with build plugins",
        str(len(plugin_packages)),
        format_percentage(len(plugin_packages), total_packages),
    )

    # Display the table
    console = rich.get_console()
    console.print(table)



from __future__ import annotations

import logging
import os
import pathlib
import platform
import re
import subprocess
import sys
import typing
from io import TextIOWrapper

from packaging.requirements import Requirement

from . import dependencies, external_commands, metrics, resolver
from .requirements_file import RequirementType
from .threading_utils import with_thread_lock

if typing.TYPE_CHECKING:
    from . import context

logger = logging.getLogger(__name__)

# Pip has no API, so parse its output looking for what it couldn't install.
# Verbose regular expressions ignore blank spaces, so we have to escape those to
# have them recognized as being part of the pattern.
_pip_missing_dependency_pattern = re.compile(
    r"""(
    Could\ not\ find\ a\ version\ that\ satisfies\ the\ requirement\ (\w+)
    |
    No\ matching\ distribution\ found\ for\ (\w+)
    |
    ResolutionImpossible  # usually when constraints prevent a match
    )""",
    flags=re.VERBOSE,
)


class MissingDependency(Exception):  # noqa: N818
    def __init__(
        self,
        ctx: context.WorkContext,
        req_type: RequirementType,
        req: Requirement | str | None,
        all_reqs: typing.Iterable[Requirement],
    ):
        self.all_reqs = all_reqs
        resolutions = []
        for r in all_reqs:
            try:
                _, version = resolver.resolve(
                    ctx=ctx,
                    req=r,
                    sdist_server_url=resolver.PYPI_SERVER_URL,
                    req_type=req_type,
                )
            except Exception as err:
                resolutions.append(f"{r} -> {err}")
            else:
                resolutions.append(f"{r} -> {version}")
        formatted_reqs = "\n".join(resolutions)
        if req:
            msg = (
                f"Failed to install {req_type} dependency {req}. "
                f"Check all {req_type} dependencies:\n{formatted_reqs}"
            )
        else:
            msg = (
                f"Failed to install {req_type} dependency. "
                f"Check all {req_type} dependencies:\n{formatted_reqs}"
            )
        super().__init__(f"\n{'*' * 40}\n{msg}\n{'*' * 40}\n")


class BuildEnvironment:
    "Wrapper for a virtualenv used for build isolation."

    def __init__(
        self,
        ctx: context.WorkContext,
        parent_dir: pathlib.Path,
    ):
        self._ctx = ctx
        self.path = parent_dir.absolute() / f"build-{platform.python_version()}"
        self._createenv()

    @property
    def python(self) -> pathlib.Path:
        """Path to Python interpreter in virtual env"""
        return self.path / "bin" / "python3"

    def get_venv_environ(
        self, template_env: dict[str, str] | None = None
    ) -> dict[str, str]:
        """Add virtual env to extra environ

        The method activates the virtualenv

        1. Put the build environment at the front of the :envvar:`PATH`
           to ensure any build tools are picked up from there and not global
           versions. If the caller has already set a path, start there.
        2. Set :envvar:`VIRTUAL_ENV` so tools looking for that (for example,
           maturin) find it.
        """
        venv_environ: dict[str, str] = {
            "VIRTUAL_ENV": str(self.path),
        }

        # pre-pend virtualenv's bin
        # 1) template_env PATH, 2) process' PATH, 3) default: "/usr/bin"
        envpath: str | None = None
        if template_env:
            envpath = template_env.get("PATH")
        if envpath is None:
            envpath = os.environ.get("PATH", "/usr/bin")
        envpath_list = envpath.split(os.pathsep)
        venv_bin = str(self.path / "bin")
        if envpath_list[0] != venv_bin:
            envpath_list.insert(0, venv_bin)
        venv_environ["PATH"] = os.pathsep.join(envpath_list)
        return venv_environ

    def run(
        self,
        cmd: typing.Sequence[str],
        *,
        cwd: str | None = None,
        extra_environ: dict[str, str] | None = None,
        network_isolation: bool | None = None,
        log_filename: str | None = None,
        stdin: TextIOWrapper | None = None,
    ) -> str:
        """Run command in a virtual environment

        `network_isolation` defaults to context setting.
        """
        extra_environ = extra_environ.copy() if extra_environ else {}
        extra_environ.update(self.get_venv_environ(template_env=extra_environ))

        # default from context
        if network_isolation is None:
            network_isolation = self._ctx.network_isolation
        if network_isolation:
            # Build Rust dependencies without network access
            extra_environ.setdefault("CARGO_NET_OFFLINE", "true")

        return external_commands.run(
            cmd,
            cwd=cwd,
            extra_environ=extra_environ,
            network_isolation=network_isolation,
            log_filename=log_filename,
            stdin=stdin,
        )

    def _createenv(self) -> None:
        if self.path.exists():
            logger.info("reusing build environment in %s", self.path)
            return

        logger.debug("creating build environment in %s", self.path)
        external_commands.run(
            [
                sys.executable,
                "-m",
                "virtualenv",
                "--python",
                sys.executable,
                "--pip=bundle",
                "--setuptools=none",
                "--no-periodic-update",
                "--no-download",
                str(self.path),
            ],
            network_isolation=self._ctx.network_isolation,
        )
        logger.info("created build environment in %s", self.path)

    @property
    def _pip_install_cmd(self) -> list[str]:
        cmd = [
            str(self.python),
            "-m",
            "pip",
            "-vvv",
            "install",
            "--disable-pip-version-check",
            "--no-compile",  # don't compile byte code
            "--upgrade",
            "--only-binary",
            ":all:",
        ]
        cmd.extend(self._ctx.pip_constraint_args)
        cmd.extend(self._ctx.pip_wheel_server_args)
        return cmd

    def install(self, reqs: typing.Iterable[Requirement]) -> None:
        if not reqs:
            return
        cmd = self._pip_install_cmd
        cmd.extend(str(req) for req in reqs)
        self.run(
            cmd,
            cwd=str(self.path.parent),
            network_isolation=False,
        )
        logger.info(
            "installed dependencies %s into build environment in %s",
            reqs,
            self.path,
        )


@metrics.timeit(description="prepare build environment")
def prepare_build_environment(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    sdist_root_dir: pathlib.Path,
) -> BuildEnvironment:
    logger.info("preparing build environment")

    build_env = BuildEnvironment(
        ctx=ctx,
        parent_dir=sdist_root_dir.parent,
    )

    build_system_dependencies = dependencies.get_build_system_dependencies(
        ctx=ctx,
        req=req,
        sdist_root_dir=sdist_root_dir,
    )
    _safe_install(
        ctx=ctx,
        req=req,
        build_env=build_env,
        deps=build_system_dependencies,
        dep_req_type=RequirementType.BUILD_SYSTEM,
    )

    build_backend_dependencies = dependencies.get_build_backend_dependencies(
        ctx=ctx,
        req=req,
        sdist_root_dir=sdist_root_dir,
        build_env=build_env,
    )
    _safe_install(
        ctx=ctx,
        req=req,
        build_env=build_env,
        deps=build_backend_dependencies,
        dep_req_type=RequirementType.BUILD_BACKEND,
    )

    build_sdist_dependencies = dependencies.get_build_sdist_dependencies(
        ctx=ctx,
        req=req,
        sdist_root_dir=sdist_root_dir,
        build_env=build_env,
    )
    _safe_install(
        ctx=ctx,
        req=req,
        build_env=build_env,
        deps=build_sdist_dependencies,
        dep_req_type=RequirementType.BUILD_SDIST,
    )

    return build_env


@with_thread_lock()
def _safe_install(
    ctx: context.WorkContext,
    req: Requirement,
    build_env: BuildEnvironment,
    deps: typing.Iterable[Requirement],
    dep_req_type: RequirementType,
) -> None:
    if not deps:
        return
    logger.debug("installing %s %s", dep_req_type, deps)
    try:
        build_env.install(deps)
    except subprocess.CalledProcessError as err:
        logger.error(
            f"{req.name}: failed to install {dep_req_type} dependencies {deps}: {err}"
        )
        match = _pip_missing_dependency_pattern.search(err.output)
        if match is not None:
            req_info = match.groups()[0]
        else:
            req_info = None
        raise MissingDependency(
            ctx,
            dep_req_type,
            req_info,
            deps,
        ) from err

    logger.info("%s: installed %s requirements %s", req.name, dep_req_type, deps)



import os
import pathlib

import click
from packaging.version import Version

from . import requirements_file


class ClickPath(click.Path):
    """ClickPath that returns pathlib.Path"""

    def convert(
        self,
        value: str | os.PathLike[str],
        param: click.core.Parameter | None,
        ctx: click.core.Context | None,
    ) -> pathlib.Path:
        path = super().convert(value=value, param=param, ctx=ctx)
        if isinstance(path, bytes):
            return pathlib.Path(os.fsdecode(path))
        return pathlib.Path(path)


class PackageVersion(click.ParamType):
    """Package version type that returns a packaging version"""

    name = "package_version"

    def convert(
        self,
        value: str,
        param: click.core.Parameter | None,
        ctx: click.core.Context | None,
    ) -> Version:
        try:
            return Version(value)
        except Exception as e:
            self.fail(
                f"Invalid package version '{value}' ({e})",
                param,
                ctx,
            )


class RequirementType(click.ParamType):
    """RequirementType that returns a requirement type"""

    name = "requirement_type"

    def convert(
        self,
        value: str,
        param: click.core.Parameter | None,
        ctx: click.core.Context | None,
    ) -> requirements_file.RequirementType:
        try:
            return requirements_file.RequirementType(value)
        except Exception:
            self.fail(
                f"Invalid requirement type '{value}', allowed values are {[r.value for r in requirements_file.RequirementType]}",
                param,
                ctx,
            )



"""Based on https://src.fedoraproject.org/rpms/python-cryptography/blob/rawhide/f/vendor_rust.py"""

import os
import pathlib
import stat
import tarfile

VCS_DIRS = {".bzr", ".git", ".hg", ".svn"}


def _tar_reset(tarinfo: tarfile.TarInfo) -> tarfile.TarInfo:
    """Reset user, group, mtime, and mode to create reproducible tar"""
    tarinfo.uid = 0
    tarinfo.gid = 0
    tarinfo.uname = "root"
    tarinfo.gname = "root"
    tarinfo.mtime = 0
    if tarinfo.type == tarfile.DIRTYPE or stat.S_IMODE(tarinfo.mode) & stat.S_IXUSR:
        tarinfo.mode = 0o755
    else:
        tarinfo.mode = 0o644
    if tarinfo.pax_headers:
        raise ValueError(tarinfo.name, tarinfo.pax_headers)
    return tarinfo


def tar_reproducible(
    tar: tarfile.TarFile,
    basedir: pathlib.Path,
    prefix: pathlib.Path | None = None,
    *,
    exclude_vcs: bool = False,
) -> None:
    """Create reproducible tar file

    Add content from basedir to already opened tar. If prefix is provided, use
    it to set relative paths for the content being added.

    If ``exclude_vcs`` is True, then Bazaar, git, Mercurial, and subversion
    directories and files are excluded.
    """
    content = [str(basedir)]  # convert from pathlib.Path, if that's what we have
    for root, dirs, files in os.walk(basedir):
        if exclude_vcs:
            # modify lists in-place, so os.walk does not descent into the
            # excluded entries. git submodules have a `.git` file.
            dirs[:] = [directory for directory in dirs if directory not in VCS_DIRS]
            files[:] = [filename for filename in files if filename not in VCS_DIRS]
        for directory in dirs:
            content.append(os.path.join(root, directory))
        for filename in files:
            content.append(os.path.join(root, filename))
    content.sort()

    for fn in content:
        # Ensure that the paths in the tarfile are rooted at the prefix
        # directory, if we have one.
        arcname = fn if prefix is None else os.path.relpath(fn, prefix)
        tar.add(fn, filter=_tar_reset, recursive=False, arcname=arcname)



from . import (
    bootstrap,
    build,
    build_order,
    canonicalize,
    download_sequence,
    graph,
    lint,
    lint_requirements,
    list_overrides,
    migrate_config,
    minimize,
    server,
    stats,
    step,
)

commands = [
    bootstrap.bootstrap,
    build.build,
    build.build_sequence,
    build.build_parallel,
    build_order.build_order,
    graph.graph,
    lint.lint,
    list_overrides.list_overrides,
    migrate_config.migrate_config,
    minimize.minimize,
    stats.stats,
    step.step,
    canonicalize.canonicalize,
    download_sequence.download_sequence,
    server.wheel_server,
    migrate_config.migrate_config,
    lint_requirements.lint_requirements,
]



from importlib import metadata

from sphinx.addnodes import desc_signature
from sphinx.application import Sphinx
from sphinx.domains.python import PyFunction
from sphinx.ext.autodoc import FunctionDocumenter
from sphinx.util.typing import ExtensionMetadata

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Fromager"
copyright = "2024, Fromager Authors"
author = "Fromager Authors"
release = metadata.version("fromager")

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx_click",
    "myst_parser",
    "sphinxcontrib.autodoc_pydantic",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
]

# Recognized suffixes
source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "example"]

language = "English"

# references to Python stdlib and packaging
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "packaging": ("https://packaging.pypa.io/en/stable/", None),
    "pyproject-hooks": ("https://pyproject-hooks.readthedocs.io/en/latest/", None),
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]

suppress_warnings = [
    "ref",  # autodoc generates a lot of warnings for missing classes that we do not want to include
]
autodoc_pydantic_model_show_json = False
autodoc_pydantic_model_show_config_summary = False
autodoc_pydantic_model_show_validator_summary = False
autodoc_pydantic_model_show_validator_members = False

# autodoc_typehints_format = "fully-qualified"


class FromagerHookDocumenter(FunctionDocumenter):
    """Documenter for 'autofromagehook' directive"""

    objtype = "fromagerhook"

    def format_name(self):
        name = super().format_name()
        if name.startswith("default_"):
            name = name[8:]
        return name


class PyFromagerHook(PyFunction):
    """:py:fromagehook"""

    def handle_signature(self, sig: str, signode: desc_signature) -> tuple[str, str]:
        # hack to remove module prefix from output
        self.options["module"] = None
        return super().handle_signature(sig, signode)


def setup(app: Sphinx) -> ExtensionMetadata:
    app.setup_extension("sphinx.ext.autodoc")
    app.add_directive_to_domain("py", FromagerHookDocumenter.objtype, PyFromagerHook)
    app.add_autodocumenter(FromagerHookDocumenter)

    return {"version": release, "parallel_read_safe": True}



import click

from fromager import context, server


@click.command()
@click.option(
    "-p",
    "--port",
    type=int,
    default=8080,
    help="the port to listen on",
)
@click.option(
    "-i",
    "--ip",
    "--address",
    "address",
    default="localhost",
    help="the address to listen on, defaults to localhost",
)
@click.pass_obj
def wheel_server(
    wkctx: context.WorkContext,
    port: int,
    address: str,
) -> None:
    "Start a web server to serve the local wheels-repo"
    server.update_wheel_mirror(wkctx)
    t = server.run_wheel_server(
        wkctx,
        address=address,
        port=port,
    )
    print(f"Listening on {wkctx.wheel_server_url}")
    t.join()



import inspect
import logging
import pathlib
import typing
from importlib import metadata

from packaging.requirements import Requirement
from packaging.utils import canonicalize_name
from stevedore import extension

# An interface for reretrieving per-package information which influences
# the build process for a particular package - i.e. for a given package
# and build target, what patches should we apply, what environment variables
# should we set, etc.
logger = logging.getLogger(__name__)


_mgr = None


def _get_extensions() -> extension.ExtensionManager:
    global _mgr
    if _mgr is None:
        _mgr = extension.ExtensionManager(
            namespace="fromager.project_overrides",
            invoke_on_load=False,
            on_load_failure_callback=_die_on_plugin_load_failure,
        )
    return _mgr


def _die_on_plugin_load_failure(
    mgr: extension.ExtensionManager,
    ep: extension.Extension,
    err: Exception,
) -> None:
    raise RuntimeError(f"failed to load overrides for {ep.name}") from err


def find_and_invoke(
    distname: str,
    method: str,
    default_fn: typing.Callable,
    **kwargs: typing.Any,
) -> typing.Any:
    fn = find_override_method(distname, method)
    if not fn:
        fn = default_fn

    result = invoke(fn, **kwargs)
    if fn is default_fn:
        log_method = logger.debug
    else:
        log_method = logger.info
    log_method(f"{distname}: override method {fn.__name__} returned {result}")

    return result


def invoke(fn: typing.Callable, **kwargs: typing.Any) -> typing.Any:
    sig = inspect.signature(fn)
    for arg_name in list(kwargs):
        if arg_name not in sig.parameters:
            logger.warning(
                f"{fn.__module__}.{fn.__name__} override does not take argument {arg_name}"
            )
            kwargs.pop(arg_name)
    return fn(**kwargs)


def _get_dist_info(package_name: str) -> tuple[str, str]:
    dists = metadata.packages_distributions()
    dist_names = dists.get(package_name.split(".")[0])
    if not dist_names:
        return (package_name, "unknown version")
    # package_distributions() returns a mapping of top-level package name to a
    # list of distribution names. The list will only have more than one element
    # if it is a namespace package. For now, assume we do not have that case and
    # take the first element of the list.
    dist_name = dist_names[0]
    dist_version = metadata.version(dist_name)
    return (dist_name, dist_version)


def log_overrides() -> None:
    for ext in _get_extensions():
        dist_name, dist_version = _get_dist_info(ext.module_name)
        logger.debug(
            "loaded override %r: from %s (%s %s)",
            ext.name,
            ext.module_name,
            dist_name,
            dist_version,
        )


def get_versioned_patch_directories(
    patches_dir: pathlib.Path,
    req: Requirement,
) -> typing.Iterator[pathlib.Path]:
    """
    This function will return directories that may contain patches for any version of a specific requirement.
    """
    # Get the req name as per the source_root_dir naming conventions
    override_name = pkgname_to_override_module(req.name)
    return patches_dir.glob(f"{override_name}-*")


def pkgname_to_override_module(pkgname: str) -> str:
    canonical_name = canonicalize_name(pkgname)
    module_name = canonical_name.replace("-", "_")
    return module_name


def find_override_method(distname: str, method: str) -> typing.Callable | None:
    """Given a distname and method name, look for an override implementation of the method.

    If there is no module or no method, return None.

    If the module exists and cannot be imported, propagate the exception.
    """
    distname = pkgname_to_override_module(distname)
    try:
        mod = _get_extensions()[distname].plugin
    except KeyError:
        logger.debug(
            "%s: no override module among %s",
            distname,
            _get_extensions().entry_points_names(),
        )
        return None
    if not hasattr(mod, method):
        logger.debug("%s: no %s override", distname, method)
        return None
    logger.info("%s: found %s override", distname, method)
    return typing.cast(typing.Callable, getattr(mod, method))



import functools
import itertools
import pathlib
import shutil
import typing
from unittest.mock import Mock, patch

import pytest
from packaging.requirements import Requirement

from fromager import build_environment, context, dependencies

_fromager_root = pathlib.Path(__file__).parent.parent


@pytest.mark.parametrize(
    "build_system,expected_results",
    [
        # Empty
        ({}, dependencies._DEFAULT_BACKEND),
        # Only specify requirements (pyarrow)
        (
            {"requires": ["a-dep"]},
            {
                "build-backend": "setuptools.build_meta:__legacy__",
                "backend-path": None,
                "requires": ["a-dep"],
            },
        ),
        # Specify everything
        (
            {
                "build-backend": "setuptools.build_meta:__legacy__",
                "backend-path": None,
                "requires": ["a-dep"],
            },
            {
                "build-backend": "setuptools.build_meta:__legacy__",
                "backend-path": None,
                "requires": ["a-dep"],
            },
        ),
    ],
)
def test_get_build_backend(
    build_system: dict[str, list[str]] | dict[str, str | list[str] | None],
    expected_results: dict[str, typing.Any] | dict[str, str | list[str] | None],
):
    pyproject_toml = {"build-system": build_system}
    actual = dependencies.get_build_backend(pyproject_toml)
    assert expected_results == actual


def _clean_build_artifacts(f):
    @functools.wraps(f)
    def _with_cleanup(*args, **kwds):
        try:
            f(*args, **kwds)
        finally:
            for d in itertools.chain(
                _fromager_root.glob("fromager-*.dist-info"),
                _fromager_root.glob("fromager.egg-info"),
            ):
                shutil.rmtree(d)

    return _with_cleanup


@patch("fromager.dependencies._write_requirements_file")
@_clean_build_artifacts
def test_get_build_system_dependencies(
    _: Mock, tmp_context: context.WorkContext, tmp_path: pathlib.Path
):
    pyproject_file = _fromager_root / "pyproject.toml"
    shutil.copyfile(pyproject_file, tmp_path / "pyproject.toml")

    results = dependencies.get_build_system_dependencies(
        ctx=tmp_context,
        req=Requirement("fromager"),
        sdist_root_dir=tmp_path,
    )
    names = set(r.name for r in results)
    assert names == set(["hatchling", "hatch-vcs"])


def test_get_build_system_dependencies_cached(
    tmp_context: context.WorkContext, tmp_path: pathlib.Path
):
    sdist_root_dir = tmp_path / "sdist"
    sdist_root_dir.mkdir()

    req_file = tmp_path / "build-system-requirements.txt"
    req_file.write_text("foo==1.0")
    results = dependencies.get_build_system_dependencies(
        ctx=tmp_context,
        req=Requirement("fromager"),
        sdist_root_dir=sdist_root_dir,
    )
    assert results == set([Requirement("foo==1.0")])


@patch("fromager.dependencies._write_requirements_file")
@_clean_build_artifacts
def test_get_build_backend_dependencies(
    _: Mock, tmp_context: context.WorkContext, tmp_path: pathlib.Path
):
    # We have to install the build system dependencies into the build
    # environment to get the build sdist dependencies, and we are not running
    # our own local wheel server, so use the public one.
    tmp_context.wheel_server_url = "https://pypi.org/simple"

    req = Requirement("fromager")
    build_env = build_environment.BuildEnvironment(
        ctx=tmp_context,
        parent_dir=tmp_path,
    )
    build_system_dependencies = dependencies.get_build_system_dependencies(
        ctx=tmp_context,
        req=req,
        sdist_root_dir=_fromager_root,
    )
    build_env.install(build_system_dependencies)

    results = dependencies.get_build_backend_dependencies(
        ctx=tmp_context,
        req=req,
        sdist_root_dir=_fromager_root,
        build_env=build_env,
    )
    names = set(r.name for r in results)
    assert names == set()


def test_get_build_backend_dependencies_cached(
    tmp_context: context.WorkContext, tmp_path: pathlib.Path
):
    sdist_root_dir = tmp_path / "sdist"
    sdist_root_dir.mkdir()

    req_file = tmp_path / "build-backend-requirements.txt"
    req_file.write_text("foo==1.0")

    build_env = build_environment.BuildEnvironment(
        ctx=tmp_context,
        parent_dir=tmp_path,
    )
    results = dependencies.get_build_backend_dependencies(
        ctx=tmp_context,
        req=Requirement("fromager"),
        sdist_root_dir=sdist_root_dir,
        build_env=build_env,
    )
    assert results == set([Requirement("foo==1.0")])


@patch("fromager.dependencies._write_requirements_file")
@_clean_build_artifacts
def test_get_build_sdist_dependencies(
    _: Mock, tmp_context: context.WorkContext, tmp_path: pathlib.Path
):
    # We have to install the build system dependencies into the build
    # environment to get the build sdist dependencies, and we are not running
    # our own local wheel server, so use the public one.
    tmp_context.wheel_server_url = "https://pypi.org/simple"

    req = Requirement("fromager")
    build_env = build_environment.BuildEnvironment(
        ctx=tmp_context,
        parent_dir=tmp_path,
    )
    build_system_dependencies = dependencies.get_build_system_dependencies(
        ctx=tmp_context,
        req=req,
        sdist_root_dir=_fromager_root,
    )
    build_env.install(build_system_dependencies)

    results = dependencies.get_build_sdist_dependencies(
        ctx=tmp_context,
        req=req,
        sdist_root_dir=_fromager_root,
        build_env=build_env,
    )
    names = set(r.name for r in results)
    assert names == set()


def test_get_build_sdist_dependencies_cached(
    tmp_context: context.WorkContext, tmp_path: pathlib.Path
):
    sdist_root_dir = tmp_path / "sdist"
    sdist_root_dir.mkdir()

    req_file = tmp_path / "build-sdist-requirements.txt"
    req_file.write_text("foo==1.0")

    req = Requirement("fromager")
    build_env = build_environment.BuildEnvironment(
        ctx=tmp_context,
        parent_dir=tmp_path,
    )
    results = dependencies.get_build_sdist_dependencies(
        ctx=tmp_context,
        req=req,
        sdist_root_dir=sdist_root_dir,
        build_env=build_env,
    )
    assert results == set([Requirement("foo==1.0")])



import pathlib
import textwrap
import typing

import tomlkit
from packaging.requirements import Requirement
from packaging.utils import canonicalize_name

from fromager import context, pyproject

PYPROJECT_TOML = """
[build-system]
requires = ["cmake", "setuptools>48.0", "torch>=2.3.0"]
"""

TEST_SETTINGS = {
    "pyproject_overrides": {
        "auto_build_requires": {
            "ninja": "ninja",
            "packaging": "packaging",
            "pybind11": "pybind11",
            "setuptools": "setuptools",
            "torch": "torch<2.4.0,>=2.3.1",
            "wheels": "wheels",
        },
        "remove_build_requires": [
            "cmake",
        ],
        "replace_build_requires": {"setuptools": "setuptools>=68.0.0"},
    }
}


def test_pyproject_no_toml(
    tmp_context: context.WorkContext, tmp_path: pathlib.Path
) -> None:
    req = Requirement("testproject==1.0.0")
    fixer = pyproject.PyprojectFix(
        req,
        build_dir=tmp_path,
        update_build_requires=["setuptools>=68.0.0"],
        remove_build_requires=[canonicalize_name("cmake")],
    )
    fixer.run()
    with tmp_path.joinpath("pyproject.toml").open(encoding="utf-8") as f:
        doc = tomlkit.load(f)
    assert isinstance(doc["build-system"], typing.Container)
    assert dict(doc["build-system"].items()) == {
        "requires": ["setuptools>=68.0.0"],
    }


def test_pyproject_with_toml(
    tmp_context: context.WorkContext, tmp_path: pathlib.Path
) -> None:
    tmp_path.joinpath("pyproject.toml").write_text(
        textwrap.dedent("""
            [build-system]
            build-backend = "maturin"
            requires = ["setuptools>=64", "setuptools_scm>=8", "maturin", "cmake>3.0", "torch<2.4.0,>=2.3.1"]
        """)
    )
    req = Requirement("testproject==1.0.0")
    fixer = pyproject.PyprojectFix(
        req,
        build_dir=tmp_path,
        update_build_requires=["setuptools>=68.0.0", "torch"],
        remove_build_requires=[canonicalize_name("cmake")],
    )
    fixer.run()
    with tmp_path.joinpath("pyproject.toml").open(encoding="utf-8") as f:
        doc = tomlkit.load(f)
    assert isinstance(doc["build-system"], typing.Container)
    assert dict(doc["build-system"].items()) == {
        "build-backend": "maturin",
        "requires": [
            "maturin",
            "setuptools>=68.0.0",
            "setuptools_scm>=8",
            "torch",
        ],
    }


def test_pyproject_fix(
    tmp_path: pathlib.Path,
    testdata_context: context.WorkContext,
) -> None:
    pyproject_toml = tmp_path / "python" / "pyproject.toml"
    pyproject_toml.parent.mkdir()
    pyproject_toml.write_text(PYPROJECT_TOML)

    req = Requirement("test-pkg==1.0.0")
    pyproject.apply_project_override(testdata_context, req, tmp_path)

    doc = tomlkit.loads(pyproject_toml.read_text())
    assert isinstance(doc["build-system"], typing.Container)
    assert dict(doc["build-system"].items()) == {
        "requires": [
            "setuptools>=68.0.0",
            "torch",
        ],
    }



import logging
import typing
from email.message import EmailMessage, Message
from email.parser import BytesParser
from io import BytesIO
from typing import TYPE_CHECKING
from zipfile import ZipFile

from packaging.requirements import Requirement
from packaging.utils import BuildTag, canonicalize_name
from packaging.version import Version

from .request_session import session

logger = logging.getLogger(__name__)

# fix for runtime errors caused by inheriting classes that are generic in stubs but not runtime
# https://mypy.readthedocs.io/en/latest/runtime_troubles.html#using-classes-that-are-generic-in-stubs-but-not-at-runtime
if TYPE_CHECKING:
    Metadata = Message[str, str]
else:
    Metadata = Message


class Candidate:
    def __init__(
        self,
        name: str,
        version: Version,
        url: str,
        extras: typing.Iterable[str] | None = None,
        is_sdist: bool | None = None,
        build_tag: BuildTag = (),
        metadata_url: str | None = None,
    ):
        self.name = canonicalize_name(name)
        self.version = version
        self.url = url
        self.extras = extras
        self.is_sdist = is_sdist
        self.build_tag = build_tag
        self.metadata_url = metadata_url

        self._metadata: Metadata | None = None
        self._dependencies: list[Requirement] | None = None

    def __repr__(self) -> str:
        if not self.extras:
            return f"<{self.name}=={self.version}>"
        return f"<{self.name}[{','.join(self.extras)}]=={self.version}>"

    @property
    def metadata(self) -> Metadata:
        if self._metadata is None:
            self._metadata = get_metadata_for_wheel(self.url, self.metadata_url)
        return self._metadata

    def _get_dependencies(self) -> typing.Iterable[Requirement]:
        deps = self.metadata.get_all("Requires-Dist", [])
        extras = self.extras if self.extras else [""]

        for d in deps:
            r = Requirement(d)
            if r.marker is None:
                yield r
            else:
                for e in extras:
                    if r.marker.evaluate({"extra": e}):
                        yield r

    @property
    def dependencies(self) -> list[Requirement]:
        if self._dependencies is None:
            self._dependencies = list(self._get_dependencies())
        return self._dependencies

    @property
    def requires_python(self) -> str | None:
        return self.metadata.get("Requires-Python")


def get_metadata_for_wheel(url: str, metadata_url: str | None = None) -> Metadata:
    """
    Get metadata for a wheel, supporting PEP 658 metadata endpoints.

    Args:
        url: URL of the wheel file
        metadata_url: Optional URL of the metadata file (PEP 658)

    Returns:
        Parsed metadata as a Message object
    """
    # Try PEP 658 metadata endpoint first if available
    if metadata_url:
        try:
            logger.debug(
                f"Attempting to fetch metadata from PEP 658 endpoint: {metadata_url}"
            )
            response = session.get(metadata_url)
            response.raise_for_status()

            # Parse metadata directly from the response content
            p = BytesParser()
            metadata = p.parse(BytesIO(response.content), headersonly=True)
            logger.debug(f"Successfully retrieved metadata via PEP 658 for {url}")
            return metadata

        except Exception as e:
            logger.debug(f"Failed to fetch PEP 658 metadata from {metadata_url}: {e}")
            logger.debug(
                "Falling back to downloading full wheel for metadata extraction"
            )

    # Fallback to existing method: download wheel and extract metadata
    logger.debug(f"Downloading full wheel to extract metadata: {url}")
    data = session.get(url).content
    with ZipFile(BytesIO(data)) as z:
        for n in z.namelist():
            if n.endswith(".dist-info/METADATA"):
                p = BytesParser()
                return p.parse(z.open(n), headersonly=True)

    # If we didn't find the metadata, return an empty dict
    return EmailMessage()



import pathlib
import typing

import pytest
from click.testing import CliRunner

from fromager import context, packagesettings

TESTDATA_PATH = pathlib.Path(__file__).parent.absolute() / "testdata"


@pytest.fixture
def testdata_path() -> typing.Generator[pathlib.Path, None, None]:
    yield TESTDATA_PATH


@pytest.fixture
def tmp_context(tmp_path: pathlib.Path) -> context.WorkContext:
    patches_dir = tmp_path / "overrides/patches"
    variant = "cpu"
    ctx = context.WorkContext(
        active_settings=None,
        constraints_file=None,
        patches_dir=patches_dir,
        sdists_repo=tmp_path / "sdists-repo",
        wheels_repo=tmp_path / "wheels-repo",
        work_dir=tmp_path / "work-dir",
        variant=variant,
    )
    ctx.setup()
    return ctx


@pytest.fixture
def testdata_context(
    testdata_path: pathlib.Path, tmp_path: pathlib.Path
) -> context.WorkContext:
    overrides = testdata_path / "context" / "overrides"
    patches_dir = overrides / "patches"
    variant = "cpu"
    ctx = context.WorkContext(
        active_settings=packagesettings.Settings.from_files(
            settings_file=overrides / "settings.yaml",
            settings_dir=overrides / "settings",
            patches_dir=patches_dir,
            variant=variant,
            max_jobs=None,
        ),
        constraints_file=None,
        patches_dir=overrides / "patches",
        sdists_repo=tmp_path / "sdists-repo",
        wheels_repo=tmp_path / "wheels-repo",
        work_dir=tmp_path / "work-dir",
    )
    ctx.setup()
    return ctx


@pytest.fixture
def cli_runner(
    tmp_path: pathlib.Path,
) -> typing.Generator[CliRunner, None, None]:
    """Click CLI runner"""
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        yield runner



import pytest
from packaging.requirements import Requirement
from packaging.version import Version

from fromager.versionmap import VersionMap


def test_initialize():
    m = VersionMap(
        {
            "1.2": "value for 1.2",
            Version("1.3"): "value for 1.3",
            "1.0": "value for 1.0",
        }
    )
    assert list(m.versions()) == [Version("1.3"), Version("1.2"), Version("1.0")]


def test_lookup():
    m = VersionMap(
        {
            "1.2": "value for 1.2",
            Version("1.3"): "value for 1.3",
            "1.0": "value for 1.0",
        }
    )
    assert m.lookup(Requirement("pkg")) == (Version("1.3"), "value for 1.3")
    assert m.lookup(Requirement("pkg>1.0")) == (Version("1.3"), "value for 1.3")
    assert m.lookup(Requirement("pkg<1.3")) == (Version("1.2"), "value for 1.2")


def test_prerelease():
    m = VersionMap(
        {
            Version("0.4.1b0"): "value for 0.4.1b0",
            "1.2": "value for 1.2",
            Version("1.3"): "value for 1.3",
            "1.0": "value for 1.0",
            "1.5.0a0": "value for 1.5.0a0",
        }
    )
    assert m.lookup(Requirement("pkg")) == (Version("1.3"), "value for 1.3")
    assert m.lookup(Requirement("pkg>1.0")) == (Version("1.3"), "value for 1.3")
    assert m.lookup(Requirement("pkg<1.3")) == (Version("1.2"), "value for 1.2")
    assert m.lookup(Requirement("pkg"), allow_prerelease=True) == (
        Version("1.5.0a0"),
        "value for 1.5.0a0",
    )
    with pytest.raises(ValueError):
        assert (
            m.lookup(Requirement("pkg"), Requirement("pkg<1.0")) == "value for 0.4.1b"
        )
    assert m.lookup(
        Requirement("pkg"), Requirement("pkg<1.0"), allow_prerelease=True
    ) == (Version("0.4.1b0"), "value for 0.4.1b0")


def test_only_prerelease():
    m = VersionMap(
        {
            Version("0.4.1b0"): "value for 0.4.1b0",
            Version("0.6b0"): "value for 0.6b0",
        }
    )
    assert m.lookup(
        Requirement("pkg"), constraint=Requirement("pkg<0.6b"), allow_prerelease=True
    ) == (
        Version("0.4.1b0"),
        "value for 0.4.1b0",
    )


def test_with_constraint():
    m = VersionMap(
        {
            "1.2": "value for 1.2",
            Version("1.3"): "value for 1.3",
            "1.0": "value for 1.0",
        }
    )
    assert m.lookup(Requirement("pkg"), Requirement("pkg<1.3")) == (
        Version("1.2"),
        "value for 1.2",
    )
    assert m.lookup(Requirement("pkg>1.0"), Requirement("pkg==1.2")) == (
        Version("1.2"),
        "value for 1.2",
    )


def test_no_match():
    m = VersionMap(
        {
            "1.2": "value for 1.2",
            Version("1.3"): "value for 1.3",
            "1.0": "value for 1.0",
        }
    )
    with pytest.raises(ValueError):
        m.lookup(Requirement("pkg"), Requirement("pkg<1.0"))
    with pytest.raises(ValueError):
        m.lookup(Requirement("pkg>1.0"), Requirement("pkg<1.0"))



import collections
import logging
import os
import pathlib
import threading
from urllib.parse import urlparse

from packaging.requirements import Requirement
from packaging.utils import NormalizedName, canonicalize_name
from packaging.version import Version

from . import constraints, dependency_graph, packagesettings, request_session

logger = logging.getLogger(__name__)

# Map package names to (requirement type, dependency name, version)
BuildRequirements = dict[str, list[tuple[str, NormalizedName, Version, Requirement]]]
ROOT_BUILD_REQUIREMENT = canonicalize_name("", validate=False)


class WorkContext:
    def __init__(
        self,
        active_settings: packagesettings.Settings | None,
        constraints_file: str | None,
        patches_dir: pathlib.Path,
        sdists_repo: pathlib.Path,
        wheels_repo: pathlib.Path,
        work_dir: pathlib.Path,
        cleanup: bool = True,
        variant: str = "cpu",
        network_isolation: bool = False,
        max_jobs: int | None = None,
        settings_dir: pathlib.Path | None = None,
    ):
        if active_settings is None:
            active_settings = packagesettings.Settings(
                settings=packagesettings.SettingsFile(),
                package_settings=[],
                patches_dir=patches_dir,
                variant=variant,
                max_jobs=max_jobs,
            )
        self.settings = active_settings
        self.input_constraints_uri: str | None
        self.constraints = constraints.Constraints()
        if constraints_file is not None:
            self.input_constraints_uri = constraints_file
            self.constraints.load_constraints_file(constraints_file)
        else:
            self.input_constraints_uri = None
        self.sdists_repo = pathlib.Path(sdists_repo).absolute()
        self.sdists_downloads = self.sdists_repo / "downloads"
        self.sdists_builds = self.sdists_repo / "builds"
        self.wheels_repo = pathlib.Path(wheels_repo).absolute()
        self.wheels_build_base = self.wheels_repo / "build"
        self.wheels_downloads = self.wheels_repo / "downloads"
        self.wheels_prebuilt = self.wheels_repo / "prebuilt"
        self.wheel_server_dir = self.wheels_repo / "simple"
        self.work_dir = pathlib.Path(work_dir).absolute()
        self.wheel_server_url = ""
        self.logs_dir = self.work_dir / "logs"
        self.cleanup = cleanup
        self.variant = variant
        self.network_isolation = network_isolation
        self.settings_dir = settings_dir

        self._constraints_filename = self.work_dir / "constraints.txt"

        self.dependency_graph = dependency_graph.DependencyGraph()

        # storing metrics
        self.time_store: dict[str, dict[str, float]] = collections.defaultdict(
            dict[str, float]
        )
        self.time_description_store: dict[str, str] = {}

        self._parallel_builds = False

    def enable_parallel_builds(self) -> None:
        self._parallel_builds = True

    @property
    def wheels_build(self) -> pathlib.Path:
        # when parallel builds are enabled, return a path that is unique for the
        # current thread to avoid collisions when creating output files
        if self._parallel_builds:
            thread_path = self.wheels_build_base / f"{threading.get_native_id()}"
            thread_path.mkdir(parents=True, exist_ok=True)
            return thread_path
        else:
            return self.wheels_build_base

    @property
    def pip_wheel_server_args(self) -> list[str]:
        args = ["--index-url", self.wheel_server_url]
        parsed = urlparse(self.wheel_server_url)
        if parsed.scheme != "https" and parsed.hostname:
            args = args + ["--trusted-host", parsed.hostname]
        return args

    @property
    def pip_constraint_args(self) -> list[str]:
        if not self.input_constraints_uri:
            return []

        if self.input_constraints_uri.startswith(("https://", "http://", "file://")):
            path_to_constraints_file = self.work_dir / "input-constraints.txt"
            if not path_to_constraints_file.exists():
                response = request_session.session.get(self.input_constraints_uri)
                path_to_constraints_file.write_text(response.text)
        else:
            path_to_constraints_file = pathlib.Path(self.input_constraints_uri)

        path_to_constraints_file = path_to_constraints_file.absolute()
        return ["--constraint", os.fspath(path_to_constraints_file)]

    def write_to_graph_to_file(self):
        with open(self.work_dir / "graph.json", "w") as f:
            self.dependency_graph.serialize(f)

    def package_build_info(
        self, package: str | packagesettings.Package | Requirement
    ) -> packagesettings.PackageBuildInfo:
        if isinstance(package, Requirement):
            name = package.name
        else:
            name = package
        return self.settings.package_build_info(name)

    def setup(self) -> None:
        # The work dir must already exist, so don't try to create it.
        # Use os.makedirs() to create the others in case the paths
        # already exist.
        for p in [
            self.work_dir,
            self.sdists_repo,
            self.sdists_downloads,
            self.sdists_builds,
            self.wheels_repo,
            self.wheels_downloads,
            self.wheels_prebuilt,
            self.wheels_build,
            self.logs_dir,
        ]:
            if not p.exists():
                logger.debug("creating %s", p)
                p.mkdir(parents=True)



import os

from .http_retry import create_retry_session

# Enhanced retry configuration for fromager
FROMAGER_RETRY_CONFIG = {
    "total": int(os.environ.get("FROMAGER_HTTP_RETRIES", "8")),
    "backoff_factor": float(os.environ.get("FROMAGER_HTTP_BACKOFF_FACTOR", "1.5")),
    "status_forcelist": [429, 500, 502, 503, 504],
    "allowed_methods": ["GET", "PUT", "POST", "HEAD", "OPTIONS"],
    "raise_on_status": False,
}

# Create a session with enhanced retry capabilities
session = create_retry_session(
    retry_config=FROMAGER_RETRY_CONFIG,
    timeout=float(os.environ.get("FROMAGER_HTTP_TIMEOUT", "120.0")),
)



import click

from .. import overrides


@click.command()
@click.argument("dist_name", nargs=-1)
def canonicalize(dist_name: list[str]) -> None:
    """convert a package name to its canonical form for use in override paths"""
    for name in dist_name:
        print(overrides.pkgname_to_override_module(name))



import io
import pathlib
import typing
from contextlib import contextmanager
from urllib.parse import urlparse

from .request_session import session


@contextmanager
def open_file_or_url(
    path_or_url: str | pathlib.Path,
) -> typing.Generator[io.TextIOBase, typing.Any, None]:
    location = str(path_or_url)
    if location.startswith("file://"):
        location = urlparse(location).path

    if location.startswith(("https://", "http://")):
        try:
            response = session.get(location)
            response.raise_for_status()
            yield io.StringIO(response.text)
        except Exception as e:
            raise OSError(f"Failed to read from URL {location}: {e}") from e
    else:
        with open(location, "r") as f:
            yield f



import concurrent.futures
import dataclasses
import datetime
import functools
import json
import logging
import pathlib
import sys
import typing
from urllib.parse import urlparse

import click
import rich
import rich.box
from packaging.requirements import Requirement
from packaging.utils import canonicalize_name, parse_wheel_filename
from packaging.version import Version
from rich.table import Table
from rich.text import Text

from fromager import (
    build_environment,
    clickext,
    context,
    dependency_graph,
    hooks,
    metrics,
    overrides,
    progress,
    read,
    server,
    sources,
    wheels,
)

from ..log import requirement_ctxvar

logger = logging.getLogger(__name__)


@dataclasses.dataclass()
@functools.total_ordering
class BuildSequenceEntry:
    name: str
    version: Version
    prebuilt: bool
    download_url: str
    wheel_filename: pathlib.Path
    skipped: bool = False

    @staticmethod
    def dict_factory(x):
        return {
            k: str(v) if isinstance(v, pathlib.Path | Version) else v for (k, v) in x
        }

    def __lt__(self, other):
        if not isinstance(other, BuildSequenceEntry):
            return NotImplemented
        # sort by lower name and version
        return (self.name.lower(), self.version) < (other.name.lower(), other.version)


@click.command()
@click.option(
    "--wheel-server-url",
    default="",
    type=str,
    help="URL for the wheel server for builds",
)
@click.argument("dist_name")
@click.argument("dist_version", type=clickext.PackageVersion())
@click.argument("sdist_server_url")
@click.pass_obj
def build(
    wkctx: context.WorkContext,
    wheel_server_url: str,
    dist_name: str,
    dist_version: Version,
    sdist_server_url: str,
) -> None:
    """Build a single version of a single wheel

    DIST_NAME is the name of a distribution

    DIST_VERSION is the version to process

    SDIST_SERVER_URL is the URL for a PyPI-compatible package index hosting sdists

    1. Downloads the source distribution.

    2. Unpacks it and prepares the source via patching, vendoring rust
       dependencies, etc.

    3. Prepares a build environment with the build dependencies.

    4. Builds the wheel.

    Refer to the 'step' commands for scripting these stages
    separately.

    """
    wkctx.wheel_server_url = wheel_server_url
    server.start_wheel_server(wkctx)
    req = Requirement(f"{dist_name}=={dist_version}")
    token = requirement_ctxvar.set(req)
    source_url, version = sources.resolve_source(
        ctx=wkctx,
        req=req,
        sdist_server_url=sdist_server_url,
    )
    wheel_filename = _build(wkctx, version, req, source_url)
    requirement_ctxvar.reset(token)
    print(wheel_filename)


build._fromager_show_build_settings = True  # type: ignore


@click.command()
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="rebuild wheels even if they have already been built",
)
@click.option(
    "-c",
    "--cache-wheel-server-url",
    "cache_wheel_server_url",
    help="url to a wheel server from where fromager can check if it had already built the wheel",
)
@click.argument("build_order_file")
@click.pass_obj
def build_sequence(
    wkctx: context.WorkContext,
    build_order_file: str,
    force: bool,
    cache_wheel_server_url: str | None,
) -> None:
    """Build a sequence of wheels in order

    BUILD_ORDER_FILE is the build-order.json files to build

    SDIST_SERVER_URL is the URL for a PyPI-compatible package index hosting sdists

    Performs the equivalent of the 'build' command for each item in
    the build order file.

    """
    server.start_wheel_server(wkctx)
    wheel_server_urls = [wkctx.wheel_server_url]
    if cache_wheel_server_url:
        # put after local server so we always check local server first
        wheel_server_urls.append(cache_wheel_server_url)

    if force:
        logger.info(f"rebuilding all wheels even if they exist in {wheel_server_urls}")
    else:
        logger.info(
            f"skipping builds for versions of packages available at {wheel_server_urls}"
        )

    entries: list[BuildSequenceEntry] = []

    logger.info("reading build order from %s", build_order_file)
    with read.open_file_or_url(build_order_file) as f:
        for entry in progress.progress(json.load(f)):
            dist_name = entry["dist"]
            resolved_version = Version(entry["version"])
            prebuilt = entry["prebuilt"]
            source_download_url = entry["source_url"]

            # If we are building from git, use the requirement as specified so
            # we include the URL. Otherwise, create a fake requirement with the
            # name and version so we are explicitly building the expected
            # version.
            if entry["source_url_type"] == "git":
                req = Requirement(entry["req"])
            else:
                req = Requirement(f"{dist_name}=={resolved_version}")
            token = requirement_ctxvar.set(req)

            if not force:
                is_built, wheel_filename = _is_wheel_built(
                    wkctx, dist_name, resolved_version, wheel_server_urls
                )
                if is_built:
                    logger.info(
                        "%s: skipping building wheel for %s==%s since it already exists",
                        dist_name,
                        dist_name,
                        resolved_version,
                    )
                    entries.append(
                        BuildSequenceEntry(
                            dist_name,
                            resolved_version,
                            prebuilt,
                            source_download_url,
                            wheel_filename=wheel_filename,
                            skipped=True,
                        )
                    )
                    continue

            if prebuilt:
                logger.info(
                    "%s: downloading prebuilt wheel %s==%s",
                    dist_name,
                    dist_name,
                    resolved_version,
                )
                wheel_filename = wheels.download_wheel(
                    req=req,
                    wheel_url=source_download_url,
                    output_directory=wkctx.wheels_build,
                )
                hooks.run_prebuilt_wheel_hooks(
                    ctx=wkctx,
                    req=req,
                    dist_name=dist_name,
                    dist_version=str(resolved_version),
                    wheel_filename=wheel_filename,
                )
                server.update_wheel_mirror(wkctx)
                # After we update the wheel mirror, the built file has
                # moved to a new directory.
                wheel_filename = wkctx.wheels_downloads / wheel_filename.name

            else:
                logger.info(
                    "%s: building %s==%s", dist_name, dist_name, resolved_version
                )
                wheel_filename = _build(
                    wkctx, resolved_version, req, source_download_url
                )

            entries.append(
                BuildSequenceEntry(
                    dist_name,
                    resolved_version,
                    prebuilt,
                    source_download_url,
                    wheel_filename=wheel_filename,
                )
            )
            print(wheel_filename)
            requirement_ctxvar.reset(token)
    metrics.summarize(wkctx, "Building")

    _summary(wkctx, entries)


build_sequence._fromager_show_build_settings = True  # type: ignore


def _summary(ctx: context.WorkContext, entries: list[BuildSequenceEntry]) -> None:
    output: list[typing.Any] = []
    now = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%SZ")
    output.append(Text(f"Build sequence summary {now}\n"))

    built_entries = [e for e in entries if not e.skipped and not e.prebuilt]
    if built_entries:
        output.append(
            _create_table(
                built_entries,
                title="New builds",
                box=rich.box.MARKDOWN,
                title_justify="left",
            )
        )
    else:
        output.append(Text("No new builds\n"))

    prebuilt_entries = [e for e in entries if e.prebuilt]
    if prebuilt_entries:
        output.append(
            _create_table(
                prebuilt_entries,
                title="Prebuilt wheels",
                box=rich.box.MARKDOWN,
                title_justify="left",
            )
        )
    else:
        output.append(Text("No pre-built wheels\n"))

    skipped_entries = [e for e in entries if e.skipped and not e.prebuilt]
    if skipped_entries:
        output.append(
            _create_table(
                skipped_entries,
                title="Skipped existing builds",
                box=rich.box.MARKDOWN,
                title_justify="left",
            )
        )
    else:
        output.append(Text("No skipped builds\n"))

    console = rich.get_console()
    console.print(*output, sep="\n\n")

    with open(ctx.work_dir / "build-sequence-summary.md", "w", encoding="utf-8") as f:
        console = rich.console.Console(file=f, width=sys.maxsize)
        console.print(*output, sep="\n\n")

    with open(ctx.work_dir / "build-sequence-summary.json", "w", encoding="utf-8") as f:
        json.dump(
            [
                dataclasses.asdict(e, dict_factory=BuildSequenceEntry.dict_factory)
                for e in entries
            ],
            f,
        )


def _create_table(entries: list[BuildSequenceEntry], **table_kwargs) -> Table:
    table = Table(**table_kwargs)
    table.add_column("Name", justify="right", no_wrap=True)
    table.add_column("Version", no_wrap=True)
    table.add_column("Wheel", no_wrap=True)
    table.add_column("Source URL")

    platlib_count = 0

    for info in sorted(entries):
        tags = parse_wheel_filename(info.wheel_filename.name)[3]
        if any(t.platform != "any" or t.abi != "none" for t in tags):
            platlib_count += 1
        source_filename = urlparse(info.download_url).path.rsplit("/", 1)[-1]
        table.add_row(
            info.name,
            str(info.version),
            info.wheel_filename.name,
            # escape Rich markup
            rf"\[{source_filename}]({info.download_url})",
        )

    # summary
    table.add_section()
    table.add_row(
        f"total: {len(entries)}",
        None,
        f"platlib: {platlib_count}",
        None,
    )
    return table


def _build(
    wkctx: context.WorkContext,
    resolved_version: Version,
    req: Requirement,
    source_download_url: str,
) -> pathlib.Path:
    per_wheel_logger = logging.getLogger("")

    module_name = overrides.pkgname_to_override_module(req.name)
    log_filename = module_name + "_current.log"

    wheel_log = wkctx.logs_dir / log_filename

    file_handler = logging.FileHandler(str(wheel_log), mode="w")
    per_wheel_logger.addHandler(file_handler)

    source_filename = sources.download_source(
        ctx=wkctx,
        req=req,
        version=resolved_version,
        download_url=source_download_url,
    )
    logger.debug(
        "%s: saved sdist of version %s from %s to %s",
        req.name,
        resolved_version,
        source_download_url,
        source_filename,
    )

    # Prepare source
    source_root_dir = sources.prepare_source(
        ctx=wkctx, req=req, source_filename=source_filename, version=resolved_version
    )

    # Build environment
    build_env = build_environment.prepare_build_environment(
        ctx=wkctx, req=req, sdist_root_dir=source_root_dir
    )

    # Make a new source distribution, in case we patched the code.
    sdist_filename = sources.build_sdist(
        ctx=wkctx,
        req=req,
        version=resolved_version,
        sdist_root_dir=source_root_dir,
        build_env=build_env,
    )

    # Build
    wheel_filename = wheels.build_wheel(
        ctx=wkctx,
        req=req,
        sdist_root_dir=source_root_dir,
        version=resolved_version,
        build_env=build_env,
    )

    per_wheel_logger.removeHandler(file_handler)
    file_handler.close()

    new_filename = wheel_log.with_name(wheel_filename.stem + ".log")
    wheel_log.rename(new_filename)

    hooks.run_post_build_hooks(
        ctx=wkctx,
        req=req,
        dist_name=canonicalize_name(req.name),
        dist_version=str(resolved_version),
        sdist_filename=sdist_filename,
        wheel_filename=wheel_filename,
    )

    server.update_wheel_mirror(wkctx)

    # After we update the wheel mirror, the built file has
    # moved to a new directory.
    wheel_filename = wkctx.wheels_downloads / wheel_filename.name

    return wheel_filename


def _is_wheel_built(
    wkctx: context.WorkContext,
    dist_name: str,
    resolved_version: Version,
    wheel_server_urls: list[str],
) -> tuple[True, pathlib.Path] | tuple[False, None]:
    req = Requirement(f"{dist_name}=={resolved_version}")

    try:
        logger.info(f"checking if {req} was already built")
        url, _ = wheels.resolve_prebuilt_wheel(
            ctx=wkctx,
            req=req,
            wheel_server_urls=wheel_server_urls,
        )
        pbi = wkctx.package_build_info(req)
        build_tag_from_settings = pbi.build_tag(resolved_version)
        build_tag = build_tag_from_settings if build_tag_from_settings else (0, "")
        wheel_filename = urlparse(url).path.rsplit("/", 1)[-1]
        _, _, build_tag_from_name, _ = parse_wheel_filename(wheel_filename)
        existing_build_tag = build_tag_from_name if build_tag_from_name else (0, "")
        if (
            existing_build_tag[0] > build_tag[0]
            and existing_build_tag[1] == build_tag[1]
        ):
            raise ValueError(
                f"{dist_name}: changelog for version {resolved_version} is inconsistent. Found build tag {existing_build_tag} but expected {build_tag}"
            )
        is_built = existing_build_tag == build_tag
        if is_built and wkctx.wheel_server_url not in url:
            # if the found wheel was on an external server, then download it
            wheels.download_wheel(req, url, wkctx.wheels_downloads)
            server.update_wheel_mirror(wkctx)

        return is_built, pathlib.Path(wheel_filename)
    except Exception:
        logger.info(f"could not locate prebuilt wheel. Will build {req}")
        return False, None


def _build_parallel(
    wkctx: context.WorkContext,
    resolved_version: Version,
    req: Requirement,
    source_download_url: str,
) -> pathlib.Path:
    try:
        token = requirement_ctxvar.set(req)
        return _build(wkctx, resolved_version, req, source_download_url)
    finally:
        requirement_ctxvar.reset(token)


@click.command()
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="rebuild wheels even if they have already been built",
)
@click.option(
    "-c",
    "--cache-wheel-server-url",
    "cache_wheel_server_url",
    help="url to a wheel server from where fromager can check if it had already built the wheel",
)
@click.option(
    "-m",
    "--max-workers",
    type=int,
    default=None,
    help="maximum number of parallel workers to run (default: unlimited)",
)
@click.argument("graph_file")
@click.pass_obj
def build_parallel(
    wkctx: context.WorkContext,
    graph_file: str,
    force: bool,
    cache_wheel_server_url: str | None,
    max_workers: int | None,
) -> None:
    """Build wheels in parallel based on a dependency graph

    GRAPH_FILE is a graph.json file containing the dependency relationships between packages

    Performs parallel builds of wheels based on their dependency relationships.
    Packages that have no dependencies or whose dependencies are already built
    can be built concurrently. By default, all possible packages are built in
    parallel. Use --max-workers to limit the number of concurrent builds.

    """
    wkctx.enable_parallel_builds()

    server.start_wheel_server(wkctx)
    wheel_server_urls = [wkctx.wheel_server_url]
    if cache_wheel_server_url:
        # put after local server so we always check local server first
        wheel_server_urls.append(cache_wheel_server_url)

    if force:
        logger.info(f"rebuilding all wheels even if they exist in {wheel_server_urls}")
    else:
        logger.info(
            f"skipping builds for versions of packages available at {wheel_server_urls}"
        )

    # Load the dependency graph
    logger.info("reading dependency graph from %s", graph_file)
    graph = dependency_graph.DependencyGraph.from_file(graph_file)

    # Get all nodes that need to be built (excluding prebuilt ones and the root node)
    nodes_to_build = []
    for node in graph.nodes.values():
        # Skip the root node and prebuilt nodes
        if node.key == dependency_graph.ROOT or node.pre_built:
            continue
        if not force:
            is_built, wheel_filename = _is_wheel_built(
                wkctx, node.canonicalized_name, node.version, wheel_server_urls
            )
            if is_built:
                logger.info(
                    "%s: skipping building wheel for %s==%s since it already exists",
                    node.canonicalized_name,
                    node.canonicalized_name,
                    node.version,
                )
                continue
        nodes_to_build.append(node)
    logger.info("found %d packages to build", len(nodes_to_build))

    # Sort nodes by their dependencies to ensure we build in the right order
    # A node can be built when all of its build dependencies are built
    built_nodes = set()
    entries: list[BuildSequenceEntry] = []

    with progress.progress_context(total=len(nodes_to_build)) as progressbar:
        while nodes_to_build:
            # Find nodes that can be built (all build dependencies are built)
            buildable_nodes = []
            for node in nodes_to_build:
                # Get all build dependencies (build-system, build-backend, build-sdist)
                build_deps = [
                    edge.destination_node
                    for edge in node.children
                    if edge.req_type.is_build_requirement
                ]
                # A node can be built when all of its build dependencies are built
                if all(dep.key in built_nodes for dep in build_deps):
                    buildable_nodes.append(node)

            if not buildable_nodes:
                # If we can't build anything but still have nodes, we have a cycle
                remaining = [n.key for n in nodes_to_build]
                raise ValueError(f"Circular dependency detected among: {remaining}")
            logger.info(
                "ready to build: %s",
                ", ".join(n.canonicalized_name for n in buildable_nodes),
            )

            # Check if any buildable node requires exclusive build (exclusive_build == True)
            exclusive_nodes = [
                node
                for node in buildable_nodes
                if wkctx.settings.package_build_info(
                    node.canonicalized_name
                ).exclusive_build
            ]
            if exclusive_nodes:
                # Only build the first exclusive node this round
                buildable_nodes = [exclusive_nodes[0]]
                logger.info(
                    f"{exclusive_nodes[0].canonicalized_name}: requires exclusive build, running it alone this round."
                )

            # Build up to max_workers nodes concurrently (or all if max_workers is None)
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=max_workers
            ) as executor:
                futures = []
                for node in buildable_nodes:
                    req = Requirement(f"{node.canonicalized_name}=={node.version}")
                    futures.append(
                        executor.submit(
                            _build_parallel,
                            wkctx,
                            node.version,
                            req,
                            node.download_url,
                        )
                    )

                # Wait for all builds to complete
                for node, future in zip(buildable_nodes, futures, strict=True):
                    try:
                        wheel_filename = future.result()
                        entries.append(
                            BuildSequenceEntry(
                                node.canonicalized_name,
                                node.version,
                                False,  # not prebuilt
                                node.download_url,
                                wheel_filename=wheel_filename,
                            )
                        )
                        built_nodes.add(node.key)
                        nodes_to_build.remove(node)
                        progressbar.update()
                    except Exception as e:
                        logger.error(f"Failed to build {node.key}: {e}")
                        raise

    metrics.summarize(wkctx, "Building in parallel")
    _summary(wkctx, entries)


build_parallel._fromager_show_build_settings = True  # type: ignore



import json
import logging
import pathlib
import typing

from packaging.requirements import Requirement
from packaging.utils import NormalizedName, canonicalize_name
from packaging.version import Version

from .read import open_file_or_url
from .requirements_file import RequirementType

logger = logging.getLogger(__name__)

ROOT = ""


class DependencyEdgeDict(typing.TypedDict):
    req_type: str
    req: str
    key: str


class DependencyNodeDict(typing.TypedDict):
    download_url: str
    canonicalized_name: str
    version: str
    pre_built: bool
    edges: list[DependencyEdgeDict]


class DependencyNode:
    def __init__(
        self,
        req_name: NormalizedName,
        version: Version,
        download_url: str = "",
        pre_built: bool = False,
    ) -> None:
        self.canonicalized_name = req_name
        self.version = version
        self.parents: list[DependencyEdge] = []
        self.children: list[DependencyEdge] = []
        self.key = f"{self.canonicalized_name}=={version}"
        self.download_url = download_url
        self.pre_built = pre_built

    def add_child(
        self,
        child: "DependencyNode",
        req: Requirement,
        req_type: RequirementType,
    ):
        current_to_child_edge = DependencyEdge(
            req=req, req_type=req_type, destination_node=child
        )
        self.children.append(current_to_child_edge)
        child_to_current_edge = DependencyEdge(
            req=req, req_type=req_type, destination_node=self
        )
        # will create a cyclic dependency in memory, which will make it harder to garbage collect
        # not an issue for fromager since it is used as a short-lived process
        child.parents.append(child_to_current_edge)

    def get_incoming_install_edges(self) -> list["DependencyEdge"]:
        return [
            edge for edge in self.parents if edge.req_type == RequirementType.INSTALL
        ]

    def to_dict(self) -> DependencyNodeDict:
        return {
            "download_url": self.download_url,
            "pre_built": self.pre_built,
            "version": str(self.version),
            "canonicalized_name": str(self.canonicalized_name),
            "edges": [edge.to_dict() for edge in self.children],
        }

    def get_outgoing_edges(
        self, req_name: str, req_type: RequirementType
    ) -> list["DependencyEdge"]:
        return [
            edge
            for edge in self.children
            if canonicalize_name(edge.req.name) == canonicalize_name(req_name)
            and edge.req_type == req_type
        ]

    @classmethod
    def construct_root_node(cls) -> "DependencyNode":
        root = cls(
            canonicalize_name(ROOT), Version("0")
        )  # version doesn't really matter for root
        root.key = ROOT  # root has a special key type
        return root


class DependencyEdge:
    def __init__(
        self,
        destination_node: DependencyNode,
        req_type: RequirementType,
        req: Requirement,
    ) -> None:
        self.req_type = req_type
        self.req = req
        self.destination_node = destination_node

    def to_dict(self) -> DependencyEdgeDict:
        return {
            "key": self.destination_node.key,
            "req_type": str(self.req_type),
            "req": str(self.req),
        }


class DependencyGraph:
    def __init__(self) -> None:
        self.nodes: dict[str, DependencyNode] = {}
        root = DependencyNode.construct_root_node()
        self.nodes[ROOT] = root

    @classmethod
    def from_file(
        cls,
        graph_file: pathlib.Path | str,
    ) -> "DependencyGraph":
        with open_file_or_url(graph_file) as f:
            # TODO: add JSON validation to ensure it is a parsable graph json
            raw_graph = typing.cast(dict[str, dict], json.load(f))
            return cls.from_dict(raw_graph)

    @classmethod
    def from_dict(
        cls,
        graph_dict: dict[str, dict[str, typing.Any]],
    ) -> "DependencyGraph":
        graph = cls()
        stack = [ROOT]
        visited = set()
        while stack:
            curr_key = stack.pop()
            if curr_key in visited:
                continue
            node_dict = typing.cast(DependencyNodeDict, graph_dict[curr_key])
            parent_name = parent_version = None
            if curr_key != ROOT:
                parent_name = canonicalize_name(node_dict["canonicalized_name"])
                parent_version = Version(node_dict["version"])
            for raw_edge in node_dict["edges"]:
                edge_dict = typing.cast(DependencyEdgeDict, raw_edge)
                destination_node_dict = typing.cast(
                    DependencyNodeDict, graph_dict[edge_dict["key"]]
                )
                graph.add_dependency(
                    parent_name=parent_name,
                    parent_version=parent_version,
                    req_type=RequirementType(edge_dict["req_type"]),
                    req=Requirement(edge_dict["req"]),
                    req_version=Version(destination_node_dict["version"]),
                    download_url=destination_node_dict["download_url"],
                    pre_built=destination_node_dict["pre_built"],
                )
                stack.append(edge_dict["key"])
            visited.add(curr_key)
        return graph

    def _to_dict(self):
        raw_graph = {}
        stack = [self.nodes[ROOT]]
        visited = set()
        while stack:
            node = stack.pop()
            raw_graph[node.key] = node.to_dict()
            for child in node.children:
                if child.destination_node.key not in visited:
                    stack.append(child.destination_node)
            visited.add(node.key)
        return raw_graph

    def serialize(self, file_handle: typing.TextIO):
        raw_graph = self._to_dict()
        json.dump(raw_graph, file_handle, indent=2, default=str)

    def _add_node(
        self,
        req_name: NormalizedName,
        version: Version,
        download_url: str,
        pre_built: bool,
    ):
        new_node = DependencyNode(
            req_name=req_name,
            version=version,
            download_url=download_url,
            pre_built=pre_built,
        )
        # check if a node with that key already exists. if it does then use that
        node = self.nodes.get(new_node.key, new_node)
        # store the node in case we are using new_node
        self.nodes[node.key] = node
        return node

    def add_dependency(
        self,
        parent_name: NormalizedName | None,
        parent_version: Version | None,
        req_type: RequirementType,
        req: Requirement,
        req_version: Version,
        download_url: str = "",
        pre_built: bool = False,
    ) -> None:
        logger.debug(
            "recording %s %s dependency %s -> %s %s",
            req_type,
            parent_name if parent_name else f"({RequirementType.TOP_LEVEL})",
            f"=={parent_version}" if parent_version else "",
            req.name,
            req_version,
        )

        node = self._add_node(
            req_name=canonicalize_name(req.name),
            version=req_version,
            download_url=download_url,
            pre_built=pre_built,
        )

        parent_key = ROOT if parent_name is None else f"{parent_name}=={parent_version}"
        if parent_key not in self.nodes:
            raise ValueError(
                f"Trying to add {node.key} to parent {parent_key} but {parent_key} does not exist"
            )

        self.nodes[parent_key].add_child(node, req=req, req_type=req_type)

    def get_dependency_edges(
        self, match_dep_types: list[RequirementType] | None = None
    ) -> typing.Iterable[DependencyEdge]:
        visited = set()
        for edge in self._depth_first_traversal(
            self.nodes[ROOT].children,
            set(),
            match_dep_types=match_dep_types,
        ):
            if edge.destination_node.key not in visited:
                yield edge
                visited.add(edge.destination_node.key)

    def get_install_dependencies(self) -> typing.Iterable[DependencyNode]:
        for edge in self.get_dependency_edges(
            match_dep_types=[RequirementType.INSTALL, RequirementType.TOP_LEVEL]
        ):
            yield edge.destination_node

    def get_nodes_by_name(self, req_name: str | None) -> list[DependencyNode]:
        if not req_name:
            return [self.nodes[ROOT]]
        return [
            node
            for node in self.get_all_nodes()
            if node.canonicalized_name == canonicalize_name(req_name)
        ]

    def get_root_node(self) -> DependencyNode:
        return self.nodes[ROOT]

    def get_all_nodes(self) -> typing.Iterable[DependencyNode]:
        return self.nodes.values()

    def get_install_dependency_versions(
        self,
    ) -> dict[NormalizedName, list[DependencyNode]]:
        all_versions: dict[NormalizedName, list[DependencyNode]] = {}
        for node in self.get_install_dependencies():
            all_versions.setdefault(node.canonicalized_name, []).append(node)
        return all_versions

    def _depth_first_traversal(
        self,
        start_edges: list[DependencyEdge],
        visited: set[str],
        match_dep_types: list[RequirementType] | None = None,
    ) -> typing.Iterable[DependencyEdge]:
        for edge in start_edges:
            if edge.destination_node.key in visited:
                continue
            if match_dep_types and edge.req_type not in match_dep_types:
                continue
            visited.add(edge.destination_node.key)
            yield edge
            yield from self._depth_first_traversal(
                edge.destination_node.children, visited, match_dep_types
            )



import logging
import pathlib
from urllib.parse import urlparse

from packaging.requirements import Requirement

from fromager import context, external_commands

logger = logging.getLogger(__name__)


def git_clone(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    output_dir: pathlib.Path,
    repo_url: str,
    tag: str | None = None,
    ref: str | None = None,
    submodules: bool | list[str] = False,
) -> pathlib.Path:
    """Clone a git repository"""
    if tag is not None and ref is not None:
        raise ValueError("tag and ref are mutually exclusive")

    # Create a clean URL without any credentials for logging
    parsed_url = urlparse(repo_url)
    clean_url = parsed_url._replace(netloc=parsed_url.hostname or "").geturl()
    logger.info(
        "%s: cloning %s, tag %r, ref %r, submodules %r, into %s",
        req.name,
        clean_url,
        tag,
        ref,
        submodules,
        output_dir,
    )
    cmd: list[str] = ["git", "clone"]
    if tag is not None:
        # --branch works with branches and tags, but not with commits
        cmd.extend(["--branch", tag, "--depth", "1"])
    if submodules:
        if isinstance(submodules, list):
            for pathspec in submodules:
                cmd.append(f"--recurse-submodules={pathspec}")
        else:
            # all submodules
            cmd.append("--recurse-submodules")
        if tag is not None:
            cmd.append("--shallow-submodules")
    cmd.extend([repo_url, str(output_dir)])
    external_commands.run(cmd, network_isolation=False)

    # --branch only works with names, so we have to checkout the reference we
    # actually want if it is not a name
    if ref is not None:
        external_commands.run(
            ["git", "checkout", "--recurse-submodules", "--force", ref],
            cwd=str(output_dir),
            network_isolation=False,
        )

    return output_dir



import json
import pathlib
import tempfile

from click.testing import CliRunner

from fromager.__main__ import main as fromager


def test_stats_command() -> None:
    """Test the stats command with sample data"""

    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = pathlib.Path(tmp_dir)

        # Create sample requirements.txt
        requirements_file = tmp_path / "requirements.txt"
        requirements_file.write_text("""
# Test requirements file
requests==2.28.1
click==8.1.3
# Comment line should be ignored
rich==12.6.0
""")

        # Create sample graph.json
        graph_file = tmp_path / "graph.json"
        # Create a graph in the format expected by DependencyGraph
        graph_data = {
            "": {
                "download_url": "",
                "pre_built": False,
                "version": "0",
                "canonicalized_name": "",
                "edges": [
                    {
                        "key": "requests==2.28.1",
                        "req_type": "toplevel",
                        "req": "requests==2.28.1",
                    },
                    {
                        "key": "click==8.1.3",
                        "req_type": "toplevel",
                        "req": "click==8.1.3",
                    },
                    {
                        "key": "rich==12.6.0",
                        "req_type": "toplevel",
                        "req": "rich==12.6.0",
                    },
                ],
            },
            "requests==2.28.1": {
                "download_url": "https://pypi.org/simple/requests/requests-2.28.1.tar.gz",
                "pre_built": False,
                "version": "2.28.1",
                "canonicalized_name": "requests",
                "edges": [
                    {
                        "key": "urllib3==1.26.12",
                        "req_type": "install",
                        "req": "urllib3>=1.21.1,<1.27",
                    }
                ],
            },
            "click==8.1.3": {
                "download_url": "https://pypi.org/simple/click/click-8.1.3-py3-none-any.whl",
                "pre_built": True,
                "version": "8.1.3",
                "canonicalized_name": "click",
                "edges": [],
            },
            "rich==12.6.0": {
                "download_url": "https://pypi.org/simple/rich/rich-12.6.0.tar.gz",
                "pre_built": False,
                "version": "12.6.0",
                "canonicalized_name": "rich",
                "edges": [],
            },
            "urllib3==1.26.12": {
                "download_url": "https://pypi.org/simple/urllib3/urllib3-1.26.12.tar.gz",
                "pre_built": False,
                "version": "1.26.12",
                "canonicalized_name": "urllib3",
                "edges": [],
            },
        }
        graph_file.write_text(json.dumps(graph_data))

        # Ensure files exist
        assert requirements_file.exists(), "Requirements file should exist"
        assert graph_file.exists(), "Graph file should exist"

        # Create work directory structure
        work_dir = tmp_path / "work"
        work_dir.mkdir()
        settings_dir = tmp_path / "settings"
        settings_dir.mkdir()
        patches_dir = tmp_path / "patches"
        patches_dir.mkdir()

        runner = CliRunner()
        result = runner.invoke(
            fromager,
            [
                "--work-dir",
                str(work_dir),
                "--settings-dir",
                str(settings_dir),
                "--patches-dir",
                str(patches_dir),
                "stats",
                str(requirements_file),
                str(graph_file),
            ],
        )

        # Check that command runs successfully
        assert result.exit_code == 0, f"Command failed with output: {result.output}"

        # Check that output contains expected information
        assert "Build Statistics" in result.output
        assert "Requirements in requirements.txt" in result.output
        assert "Unique packages in build" in result.output
        assert "Pre-built packages" in result.output

        assert "3" in result.output  # 3 requirements
        assert "4" in result.output  # 4 unique packages


def test_stats_command_nonexistent_files() -> None:
    """Test that the stats command fails gracefully with nonexistent files"""

    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = pathlib.Path(tmp_dir)

        # Create work directory structure
        work_dir = tmp_path / "work"
        work_dir.mkdir()
        settings_dir = tmp_path / "settings"
        settings_dir.mkdir()
        patches_dir = tmp_path / "patches"
        patches_dir.mkdir()

        # Try with nonexistent files
        nonexistent_req = tmp_path / "nonexistent_requirements.txt"
        nonexistent_graph = tmp_path / "nonexistent_graph.json"

        runner = CliRunner()
        result = runner.invoke(
            fromager,
            [
                "--work-dir",
                str(work_dir),
                "--settings-dir",
                str(settings_dir),
                "--patches-dir",
                str(patches_dir),
                "stats",
                str(nonexistent_req),
                str(nonexistent_graph),
            ],
        )

        # Should fail because files don't exist
        assert result.exit_code != 0, "Command should fail with nonexistent files"



import logging

from fromager import external_commands

logger = logging.getLogger(__name__)


def build_wheel(ctx, build_env, extra_environ, req, sdist_root_dir, version):
    # flit_core is a basic build system dependency for several
    # packages. It is capable of building its own wheels, so we use the
    # bootstrapping instructions to do that and put the wheel in the
    # local server directory for reuse when building other packages via
    # 'pip wheel'.
    #
    # https://flit.pypa.io/en/stable/bootstrap.html
    logger.info('using override to build flit_core wheel in %s', sdist_root_dir)
    external_commands.run(
        [build_env.python, '-m', 'flit_core.wheel',
         '--outdir', ctx.wheels_build],
        cwd=sdist_root_dir,
        extra_environ=extra_environ,
    )



import logging
import pathlib
import typing

import click
import yaml
from packaging.utils import NormalizedName, canonicalize_name

from fromager import clickext, context, overrides, packagesettings

logger = logging.getLogger(__name__)

PkgEnv = dict[NormalizedName, packagesettings.EnvVars]
VariantPkgEnv = dict[
    NormalizedName, dict[packagesettings.Variant, packagesettings.EnvVars]
]
PrebuiltMap = dict[NormalizedName, set[packagesettings.Variant]]


def _parse_envfile(envfile: pathlib.Path) -> packagesettings.EnvVars:
    """Parse a pkg.env file from Fromager 0.27"""
    env: packagesettings.EnvVars = {}
    with envfile.open(encoding="utf-8") as f:
        for line in f:
            line = line.strip()
            if not line or line.startswith("#"):
                continue
            key, _, value = line.partition("=")
            key = packagesettings.EnvKey(key.strip())
            value = value.strip()
            if value[0] == value[-1] and (value[0] == '"' or value[0] == "'"):
                value = value[1:-1]
            env[key] = value
    return env


def _load_envfiles(envs_dir: pathlib.Path) -> tuple[PkgEnv, VariantPkgEnv]:
    """Load default and variant env files from Fromager 0.27"""
    default_env: PkgEnv = {}
    for filename in envs_dir.glob("*.env"):
        pkg = canonicalize_name(filename.stem)
        default_env[pkg] = _parse_envfile(filename)

    variants_env: VariantPkgEnv = {}
    for filename in envs_dir.glob("*/*.env"):
        variant = packagesettings.Variant(filename.parent.name)
        pkg = canonicalize_name(filename.stem)

        variants_env.setdefault(pkg, {})[variant] = _parse_envfile(filename)

    return default_env, variants_env


def _migrate_package_envfiles(
    envs_dir: pathlib.Path,
    settings_file: pathlib.Path,
    output_dir: pathlib.Path,
) -> None:
    """Migrate settings.yaml of Fromager 0.27 and earlier

    - "packages" section in setting.yaml
    - "pre_built" section in settings.yaml
    - separate env files in `env_dir` (<name>.yaml, <variant>/<name>.yaml)
    """
    with settings_file.open(encoding="utf-8") as f:
        settings_data: dict[str, typing.Any] = yaml.safe_load(f)

    settings_pkgs: dict[NormalizedName, typing.Any] = {
        canonicalize_name(name): value
        for name, value in settings_data.get("packages", {}).items()
    }

    pre_built: PrebuiltMap = {}
    for variantname, entries in settings_data.get("pre_built", {}).items():
        if not entries:
            continue
        variant = packagesettings.Variant(variantname)
        for pkgname in entries:
            name = NormalizedName(pkgname)
            pre_built.setdefault(name, set()).add(variant)

    default_env, variants_env = _load_envfiles(envs_dir)

    # packages with configuration, pre_built setting, or env files
    pkg_names: set[NormalizedName] = set()
    pkg_names.update(canonicalize_name(pkg) for pkg in settings_pkgs)
    pkg_names.update(default_env)
    pkg_names.update(variants_env)
    pkg_names.update(pre_built)

    package_settings = []
    for name in sorted(pkg_names):
        values: dict[str, typing.Any] = {"variants": {}}
        if name in settings_pkgs:
            values.update(settings_pkgs[name])
        if name in default_env:
            values["env"] = default_env[name]
        if name in variants_env:
            for variant, envvars in variants_env[name].items():
                values["variants"][variant] = {"env": envvars}
        if name in pre_built:
            for variant in pre_built[name]:
                values["variants"].setdefault(variant, {})["pre_built"] = True
        ps = packagesettings.PackageSettings.from_mapping(
            name, values, source=None, has_config=True
        )
        package_settings.append(ps)

    output_dir.mkdir(exist_ok=True)
    for ps in package_settings:
        dump = ps.serialize()
        override_name = overrides.pkgname_to_override_module(ps.name)
        filename = output_dir / f"{override_name}.yaml"
        logger.info("Creating config file %s", filename)
        filename.write_text(yaml.dump(dump), encoding="utf-8")


@click.command()
# --envs-dir will be removed from fromager base command
@click.option(
    "--envs-dir",
    type=clickext.ClickPath(),
    help="location of old environment override files",
    required=True,
)
@click.option(
    "--settings-file",
    type=clickext.ClickPath(),
    help="location of the old application settings file",
    required=True,
)
@click.option(
    "--output-dir",
    type=clickext.ClickPath(),
    help="location to write per-package settings",
    required=True,
)
@click.pass_obj
def migrate_config(
    wkctx: context.WorkContext,
    envs_dir: pathlib.Path,
    settings_file: pathlib.Path,
    output_dir: pathlib.Path,
) -> None:
    """Migrate Fromager 0.27 config to new format"""
    _migrate_package_envfiles(
        envs_dir=envs_dir,
        settings_file=settings_file,
        output_dir=output_dir,
    )



import logging
import typing

import click
from packaging.requirements import Requirement
from packaging.utils import NormalizedName
from packaging.version import Version

from fromager.dependency_graph import DependencyNode

from .. import (
    bootstrapper,
    context,
    dependency_graph,
    metrics,
    progress,
    requirements_file,
    server,
)
from ..log import requirement_ctxvar
from ..requirements_file import RequirementType
from .graph import find_why

# Map child_name==child_version to list of (parent_name==parent_version, Requirement)
ReverseRequirements = dict[str, list[tuple[str, Requirement]]]

logger = logging.getLogger(__name__)


def _get_requirements_from_args(
    toplevel: typing.Iterable[str],
    req_files: typing.Iterable[str],
) -> list[Requirement]:
    parsed_req: list[str] = []
    parsed_req.extend(toplevel)
    for filename in req_files:
        parsed_req.extend(requirements_file.parse_requirements_file(filename))
    to_build: list[Requirement] = []
    for dep in parsed_req:
        req = Requirement(dep)
        # If we're given a requirements file as input, we might be iterating over a
        # list of requirements with marker expressions that limit their use to
        # specific platforms or python versions. Evaluate the markers to filter out
        # anything we shouldn't build. Only apply the filter to toplevel
        # requirements (items without a why list leading up to them) because other
        # dependencies are already filtered based on their markers in the context of
        # their parent, so they include values like the parent's extras settings.
        if not requirements_file.evaluate_marker(req, req):
            logger.info(
                f"ignoring {requirements_file.RequirementType.TOP_LEVEL} dependency {req} because of its marker expression"
            )
        else:
            to_build.append(req)
    return to_build


@click.command()
@click.option(
    "-r",
    "--requirements-file",
    "requirements_files",
    multiple=True,
    type=str,
    help="pip requirements file",
)
@click.option(
    "-p",
    "--previous-bootstrap-file",
    "previous_bootstrap_file",
    type=str,
    help="graph file produced from a previous bootstrap",
)
@click.option(
    "-c",
    "--cache-wheel-server-url",
    "cache_wheel_server_url",
    help="url to a wheel server from where fromager can download the wheels that it has built before",
)
@click.option(
    "--sdist-only/--full-build",
    "sdist_only",
    default=False,
    help=(
        "--sdist-only (fast mode) does not build missing wheels unless they "
        "are build requirements. --full-build (default) builds all missing "
        "wheels."
    ),
)
@click.option(
    "--skip-constraints",
    "skip_constraints",
    is_flag=True,
    default=False,
    help="Skip generating constraints.txt file to allow building collections with conflicting versions",
)
@click.argument("toplevel", nargs=-1)
@click.pass_obj
def bootstrap(
    wkctx: context.WorkContext,
    requirements_files: list[str],
    previous_bootstrap_file: str | None,
    cache_wheel_server_url: str | None,
    sdist_only: bool,
    skip_constraints: bool,
    toplevel: list[str],
) -> None:
    """Compute and build the dependencies of a set of requirements recursively

    TOPLEVEL is a requirements specification, including a package name
    and optional version constraints.

    """
    logger.info(f"cache wheel server url: {cache_wheel_server_url}")

    to_build = _get_requirements_from_args(toplevel, requirements_files)
    if not to_build:
        raise RuntimeError(
            "Pass a requirement specificiation or use -r to pass a requirements file"
        )
    logger.info("bootstrapping %r variant of %s", wkctx.variant, to_build)

    if previous_bootstrap_file:
        logger.info("reading previous bootstrap data from %s", previous_bootstrap_file)
        prev_graph = dependency_graph.DependencyGraph.from_file(previous_bootstrap_file)
    else:
        logger.info("no previous bootstrap data")
        prev_graph = None

    if sdist_only:
        logger.info("sdist-only (fast mode), getting metadata from sdists")
    else:
        logger.info("build all missing wheels")

    pre_built = wkctx.settings.list_pre_built()
    if pre_built:
        logger.info("treating %s as pre-built wheels", sorted(pre_built))

    server.start_wheel_server(wkctx)

    with progress.progress_context(total=len(to_build * 2)) as progressbar:
        bt = bootstrapper.Bootstrapper(
            wkctx,
            progressbar,
            prev_graph,
            cache_wheel_server_url,
            sdist_only=sdist_only,
        )

        # we need to resolve all the top level dependencies before we start bootstrapping.
        # this is to ensure that if we are using an older bootstrap to resolve packages
        # we are able to upgrade a package anywhere in the dependency tree if it is mentioned
        # in the toplevel without having to fall back to history
        logger.info("resolving top-level dependencies before building")
        for req in to_build:
            token = requirement_ctxvar.set(req)
            pbi = wkctx.package_build_info(req)
            if pbi.pre_built:
                source_url, version = bt.resolve_version(
                    req=req,
                    req_type=RequirementType.TOP_LEVEL,
                )
            else:
                source_url, version = bt.resolve_version(
                    req=req,
                    req_type=RequirementType.TOP_LEVEL,
                )
            logger.info("%s resolves to %s", req, version)
            wkctx.dependency_graph.add_dependency(
                parent_name=None,
                parent_version=None,
                req_type=requirements_file.RequirementType.TOP_LEVEL,
                req=req,
                req_version=version,
                download_url=source_url,
                pre_built=pbi.pre_built,
            )
            requirement_ctxvar.reset(token)

        for req in to_build:
            token = requirement_ctxvar.set(req)
            bt.bootstrap(req, requirements_file.RequirementType.TOP_LEVEL)
            progressbar.update()
            requirement_ctxvar.reset(token)

    # If we put pre-built wheels in the downloads directory, we should
    # remove them so we can treat that directory as a source of wheels
    # to upload to an index.
    for prebuilt_wheel in wkctx.wheels_prebuilt.glob("*.whl"):
        filename = wkctx.wheels_downloads / prebuilt_wheel.name
        if filename.exists():
            logger.info(
                f"removing prebuilt wheel {prebuilt_wheel.name} from download cache"
            )
            filename.unlink()

    constraints_filename = wkctx.work_dir / "constraints.txt"
    if skip_constraints:
        logger.info("skipping constraints.txt generation as requested")
    else:
        logger.info(f"writing installation dependencies to {constraints_filename}")
        with open(constraints_filename, "w") as f:
            if not write_constraints_file(graph=wkctx.dependency_graph, output=f):
                raise ValueError(
                    f"Could not produce a pip compatible constraints file. Please review {constraints_filename} for more details"
                )

    metrics.summarize(wkctx, "Bootstrapping")


def write_constraints_file(
    graph: dependency_graph.DependencyGraph,
    output: typing.TextIO,
) -> bool:
    # Look for potential conflicts by tracking how many different versions of
    # each package are needed.
    conflicts: dict[NormalizedName, list[DependencyNode]] = (
        graph.get_install_dependency_versions()
    )
    ret = True
    conflicting_deps: set[NormalizedName] = set()

    # Map for already resolved versions for a given dependency Eg: {"a": "0.4"}
    resolved: dict[NormalizedName, Version] = {}

    # List of unresolved dependencies
    unresolved_dependencies: list[tuple[NormalizedName, list[DependencyNode]]] = sorted(
        conflicts.items()
    )

    dep_name: NormalizedName

    # Loop over dependencies and resolve dependencies with single version first. This will shrink the unresolved_dependencies to begin with.
    for dep_name, nodes in unresolved_dependencies[:]:
        versions: list[Version] = [node.version for node in nodes]
        if len(versions) == 0:
            # This should never happen.
            raise ValueError(f"No versions of {dep_name} supported")

        if len(versions) == 1:
            logger.debug(
                "resolving %s to %s",
                dep_name,
                versions[0],
            )
            # This is going to be the situation for most dependencies, where we
            # only have one version.
            resolved[dep_name] = versions[0]
            # Remove from unresolved dependencies list
            unresolved_dependencies.remove((dep_name, nodes))
    multiple_versions: dict[NormalizedName, list[DependencyNode]] = dict(
        unresolved_dependencies
    )

    # Below this point we have built multiple versions of the same thing, so
    # we need to try to determine if any one of those versions meets all of
    # the requirements.

    # Flag to see if something is resolved
    resolved_something: bool = True

    # Outer while loop to resolve remaining dependencies with multiple versions
    while unresolved_dependencies and resolved_something:
        logger.debug(
            "starting to resolve %s",
            [dep_name for dep_name, _ in unresolved_dependencies],
        )
        resolved_something = False
        # Make copy of the original list and loop over unresolved dependencies
        for dep_name, nodes in unresolved_dependencies[:]:
            # Track which versions can be used by which parent requirement.
            usable_versions: dict[Version, list[Version]] = {}
            # Track how many total users of a requirement (by name) there are so we
            # can tell later if any version can be used by all of them.
            user_counter: int = 0
            # Which parent requirements can use which versions of the dependency we
            # are working on?
            dep_versions: list[Version] = [node.version for node in nodes]

            # Loop over the nodes list
            for node in nodes:
                parent_edges: list[dependency_graph.DependencyEdge] = (
                    node.get_incoming_install_edges()
                )
                if not parent_edges:
                    # This is a top level dependency, so we should ensure that the
                    # resolved version is considered as a candidate.
                    usable_versions.setdefault(node.version, []).append(node.version)

                # Loop over parent_edges list
                for parent_edge in parent_edges:
                    parent_name: NormalizedName = (
                        parent_edge.destination_node.canonicalized_name
                    )
                    # Condition to select the right version.
                    # We check whether parent_name is already in resolved dict and the version associated with that
                    # is not the version of the destination node
                    if (
                        parent_name in resolved
                        and resolved[parent_name]
                        != parent_edge.destination_node.version
                    ):
                        continue
                    # Loop to find the usable versions
                    for matching_version in parent_edge.req.specifier.filter(  # type: ignore
                        dep_versions
                    ):
                        usable_versions.setdefault(matching_version, []).append(
                            parent_edge.destination_node.version
                        )
                    user_counter += 1

            # Look for one version that can be used by all the parent dependencies
            # and output that if we find it. Otherwise, include a warning and report
            # all versions so a human reading the file can make their own decision
            # about how to resolve the conflict.
            for v, users in reversed(sorted(usable_versions.items())):  # type: ignore
                logger.debug(
                    "considering %s for %s, %d of %d consumers",
                    v,
                    dep_name,
                    len(users),
                    user_counter,
                )
                if len(users) != user_counter:
                    logger.debug(
                        "%s: version %s is useable by %d of %d consumers, skipping it",
                        dep_name,
                        v,
                        len(users),
                        user_counter,
                    )
                    continue
                version_strs: list[str] = [
                    str(v) for v in reversed(sorted(dep_versions))
                ]
                logger.debug(
                    "%s: selecting %s from multiple candidates %s",
                    dep_name,
                    v,
                    version_strs,
                )
                resolved[dep_name] = v
                resolved_something = True
                try:
                    unresolved_dependencies.remove((dep_name, nodes))
                except ValueError:
                    logger.debug(
                        "%s: %s not in unresolved dependencies list, ignoring",
                        dep_name,
                        (dep_name, nodes),
                    )
                break

    # Write resolved versions to constraints file
    for dep_name, resolved_version in sorted(resolved.items()):  # type: ignore
        if dep_name in multiple_versions:
            version_strs = [
                str(node.version)
                for node in sorted(multiple_versions[dep_name], key=lambda n: n.version)
            ]
            output.write(
                f"# NOTE: fromager selected {dep_name}=={resolved_version} from: {version_strs}\n"
            )
        output.write(f"{dep_name}=={resolved_version}\n")

    # No single version could be used, so go ahead and print all the
    # versions with a warning message
    for dep_name, nodes in unresolved_dependencies:  # type: ignore
        ret = False
        logger.error("%s: no single version meets all requirements", dep_name)
        output.write(f"# ERROR: no single version of {dep_name} met all requirements\n")
        conflicting_deps.add(dep_name)
        for node in sorted(nodes, key=lambda n: n.version):
            output.write(f"{dep_name}=={node.version}\n")

    for dep_name in conflicting_deps:
        for node in graph.get_nodes_by_name(dep_name):
            find_why(
                graph=graph,
                node=node,
                max_depth=-1,
                depth=0,
                req_type=[],
            )

    return ret


bootstrap._fromager_show_build_settings = True  # type: ignore



import logging
import pathlib

import click
from packaging.requirements import Requirement
from packaging.version import Version

from .. import (
    build_environment,
    clickext,
    context,
    finders,
    server,
    sources,
    wheels,
)
from ..log import requirement_ctxvar

logger = logging.getLogger(__name__)


@click.group()
def step() -> None:
    "Step-by-step commands"
    pass


@step.command()
@click.argument("dist_name")
@click.argument("dist_version", type=clickext.PackageVersion())
@click.argument("sdist_server_url")
@click.pass_obj
def download_source_archive(
    wkctx: context.WorkContext,
    dist_name: str,
    dist_version: Version,
    sdist_server_url: str,
) -> None:
    """download the source code archive for one version of one package

    DIST_NAME is the name of a distribution

    DIST_VERSION is the version to process

    SDIST_SERVER_URL is the URL for a PyPI-compatible package index hosting sdists

    """
    req = Requirement(f"{dist_name}=={dist_version}")
    token = requirement_ctxvar.set(req)
    source_url, version = sources.resolve_source(
        ctx=wkctx, req=req, sdist_server_url=sdist_server_url
    )
    filename = sources.download_source(
        ctx=wkctx,
        req=req,
        version=version,
        download_url=source_url,
    )
    requirement_ctxvar.reset(token)
    print(filename)


@step.command()
@click.argument("dist_name")
@click.argument("dist_version", type=clickext.PackageVersion())
@click.pass_obj
def prepare_source(
    wkctx: context.WorkContext,
    dist_name: str,
    dist_version: Version,
) -> None:
    """ensure the source code is in a form ready for building a wheel

    DIST_NAME is the name of a distribution

    DIST_VERSION is the version to process

    """
    req = Requirement(f"{dist_name}=={dist_version}")
    token = requirement_ctxvar.set(req)
    sdists_downloads = pathlib.Path(wkctx.sdists_repo) / "downloads"
    source_filename = finders.find_sdist(
        wkctx, wkctx.sdists_downloads, req, str(dist_version)
    )
    if source_filename is None:
        dir_contents: list[str] = []
        for ext in ["*.tar.gz", "*.zip"]:
            dir_contents.extend(str(e) for e in wkctx.sdists_downloads.glob(ext))
        raise RuntimeError(
            f"Cannot find sdist for {req.name} version {dist_version} in {sdists_downloads} among {dir_contents}"
        )
    source_root_dir = sources.prepare_source(
        ctx=wkctx,
        req=req,
        source_filename=source_filename,
        version=dist_version,
    )
    requirement_ctxvar.reset(token)
    print(source_root_dir)


@step.command()
@click.argument("dist_name")
@click.argument("dist_version", type=clickext.PackageVersion())
@click.pass_obj
def build_sdist(
    wkctx: context.WorkContext,
    dist_name: str,
    dist_version: Version,
) -> None:
    """build a new source distribution for the package

    DIST_NAME is the name of a distribution

    DIST_VERSION is the version to process

    The source distribution is placed in the `sdists-repo/builds`
    directory.

    """
    req = Requirement(f"{dist_name}=={dist_version}")
    token = requirement_ctxvar.set(req)
    source_root_dir = _find_source_root_dir(wkctx, wkctx.work_dir, req, dist_version)
    build_env = build_environment.BuildEnvironment(
        ctx=wkctx,
        parent_dir=source_root_dir.parent,
    )
    sdist_filename = sources.build_sdist(
        ctx=wkctx,
        req=req,
        version=dist_version,
        sdist_root_dir=source_root_dir,
        build_env=build_env,
    )
    requirement_ctxvar.reset(token)
    print(sdist_filename)


def _find_source_root_dir(
    wkctx: context.WorkContext,
    work_dir: pathlib.Path,
    req: Requirement,
    dist_version: Version,
) -> pathlib.Path:
    source_root_dir = finders.find_source_dir(wkctx, work_dir, req, str(dist_version))
    if source_root_dir:
        return source_root_dir
    work_dir_contents = list(str(e) for e in work_dir.glob("*"))
    raise RuntimeError(
        f"Cannot find source directory for {req.name} version {dist_version} among {work_dir_contents}"
    )


@step.command()
@click.option(
    "--wheel-server-url",
    default="",
    type=str,
    help="URL for the wheel server for builds",
)
@click.argument("dist_name")
@click.argument("dist_version", type=clickext.PackageVersion())
@click.pass_obj
def prepare_build(
    wkctx: context.WorkContext,
    wheel_server_url: str,
    dist_name: str,
    dist_version: Version,
) -> None:
    """set up build environment to build the package

    DIST_NAME is the name of a distribution

    DIST_VERSION is the version to process

    """
    wkctx.wheel_server_url = wheel_server_url
    server.start_wheel_server(wkctx)
    req = Requirement(f"{dist_name}=={dist_version}")
    token = requirement_ctxvar.set(req)
    source_root_dir = _find_source_root_dir(wkctx, wkctx.work_dir, req, dist_version)
    build_environment.prepare_build_environment(
        ctx=wkctx, req=req, sdist_root_dir=source_root_dir
    )
    requirement_ctxvar.reset(token)


@step.command()
@click.option(
    "--wheel-server-url",
    default="",
    type=str,
    help="URL for the wheel server for builds",
)
@click.argument("dist_name")
@click.argument("dist_version", type=clickext.PackageVersion())
@click.pass_obj
def build_wheel(
    wkctx: context.WorkContext,
    wheel_server_url: str,
    dist_name: str,
    dist_version: Version,
) -> None:
    """build a wheel from prepared source

    DIST_NAME is the name of a distribution

    DIST_VERSION is the version to process

    """
    wkctx.wheel_server_url = wheel_server_url
    req = Requirement(f"{dist_name}=={dist_version}")
    token = requirement_ctxvar.set(req)
    source_root_dir = _find_source_root_dir(wkctx, wkctx.work_dir, req, dist_version)
    server.start_wheel_server(wkctx)
    build_env = build_environment.BuildEnvironment(
        ctx=wkctx,
        parent_dir=source_root_dir.parent,
    )
    wheel_filename = wheels.build_wheel(
        ctx=wkctx,
        req=req,
        sdist_root_dir=source_root_dir,
        version=dist_version,
        build_env=build_env,
    )
    requirement_ctxvar.reset(token)
    print(wheel_filename)



"""HTTP retry utilities for resilient network operations.

This module provides a robust retry mechanism for HTTP requests with exponential
backoff, jitter, and specific handling for common network failures including:
- Server timeouts (502, 503, 504)
- Rate limiting (429, GitHub API rate limits)
- Connection errors and incomplete reads
- DNS resolution failures

The retry session can be used as a drop-in replacement for the standard
requests session in most cases.
"""

from __future__ import annotations

import logging
import random
import time
import typing

import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import (
    ChunkedEncodingError,
    ConnectionError,
    RequestException,
    Timeout,
)
from urllib3.exceptions import IncompleteRead, ProtocolError
from urllib3.util.retry import Retry

logger = logging.getLogger(__name__)

# Default retry configuration
DEFAULT_RETRY_CONFIG = {
    "total": 5,
    "backoff_factor": 1.0,
    "status_forcelist": [429, 500, 502, 503, 504],
    "allowed_methods": ["GET", "PUT", "POST", "HEAD", "OPTIONS"],
    "raise_on_status": False,
}

# Exceptions that should trigger a retry
RETRYABLE_EXCEPTIONS = (
    ConnectionError,
    Timeout,
    ChunkedEncodingError,
    IncompleteRead,
    ProtocolError,
    # Add more urllib3 exceptions that are often transient
    requests.exceptions.RetryError,
    requests.exceptions.ConnectTimeout,
    requests.exceptions.ReadTimeout,
)


class RetryHTTPAdapter(HTTPAdapter):
    """HTTP adapter with enhanced retry logic and backoff."""

    def __init__(
        self,
        retry_config: dict[str, typing.Any] | None = None,
        timeout: float = 60.0,
        **kwargs: typing.Any,
    ):
        """Initialize the retry adapter.

        Args:
            retry_config: Configuration for urllib3 Retry. If None, uses DEFAULT_RETRY_CONFIG.
            timeout: Default timeout for requests in seconds.
            **kwargs: Additional arguments passed to HTTPAdapter.
        """
        self.timeout = timeout
        config = retry_config or DEFAULT_RETRY_CONFIG.copy()

        # Store configuration for use in send method with proper type handling
        backoff_factor_val = config.get("backoff_factor", 1.0)
        self.backoff_factor: float = (
            1.0
            if not isinstance(backoff_factor_val, int | float)
            else float(backoff_factor_val)
        )
        self.max_backoff: float = 60.0  # Maximum backoff time in seconds

        # Create the retry strategy with proper type handling
        total_val = config.get("total", 5)
        total = 5 if not isinstance(total_val, int | float) else int(total_val)

        backoff_val = config.get("backoff_factor", 1.0)
        backoff = (
            1.0 if not isinstance(backoff_val, int | float) else float(backoff_val)
        )

        status_list_val = config.get("status_forcelist", [429, 500, 502, 503, 504])
        status_list = (
            [429, 500, 502, 503, 504]
            if not isinstance(status_list_val, list)
            else status_list_val
        )

        methods_val = config.get(
            "allowed_methods", ["GET", "PUT", "POST", "HEAD", "OPTIONS"]
        )
        methods = (
            ["GET", "PUT", "POST", "HEAD", "OPTIONS"]
            if not isinstance(methods_val, list)
            else methods_val
        )

        raise_on_status_val = config.get("raise_on_status", False)
        raise_on_status = (
            False if not isinstance(raise_on_status_val, bool) else raise_on_status_val
        )

        retry_strategy = Retry(
            total=total,
            backoff_factor=backoff,
            status_forcelist=status_list,
            allowed_methods=methods,
            raise_on_status=raise_on_status,
        )

        super().__init__(max_retries=retry_strategy, **kwargs)

    def send(
        self,
        request: requests.PreparedRequest,
        stream: bool = False,
        timeout: float | tuple[float, float] | tuple[float, None] | None = None,
        verify: bool | str = True,
        cert: bytes | str | tuple[bytes | str, bytes | str] | None = None,
        proxies: typing.Mapping[str, str] | None = None,
        **kwargs: typing.Any,
    ) -> requests.Response:
        """Send request with enhanced error handling and retry logic."""
        if timeout is None:
            timeout = self.timeout

        send_kwargs = {
            "stream": stream,
            "timeout": timeout,
            "verify": verify,
            "cert": cert,
            "proxies": proxies,
            **kwargs,
        }

        max_attempts = getattr(self.max_retries, "total", 5) + 1

        for attempt in range(max_attempts):
            try:
                response = super().send(request, **send_kwargs)

                # Handle GitHub API rate limiting specifically
                if (
                    response.status_code == 403
                    and request.url is not None
                    and "api.github.com" in request.url
                    and "rate limit" in response.text.lower()
                ):
                    self._handle_github_rate_limit(response, attempt, max_attempts)
                    continue

                # Check for retryable HTTP status codes
                if response.status_code in {500, 502, 503, 504}:
                    if attempt >= max_attempts - 1:
                        logger.error(
                            "Request failed with status %d after %d attempts for %s",
                            response.status_code,
                            max_attempts,
                            request.url or "<unknown>",
                        )
                        return response

                    wait_time = min(
                        self.backoff_factor * (2**attempt) + random.uniform(0, 1),
                        self.max_backoff,
                    )

                    logger.warning(
                        "Request failed with status %d for %s. Retrying in %.1f seconds (attempt %d/%d)",
                        response.status_code,
                        request.url or "<unknown>",
                        wait_time,
                        attempt + 1,
                        max_attempts,
                    )
                    time.sleep(wait_time)
                    continue

                return response

            except RETRYABLE_EXCEPTIONS as e:
                self._handle_retryable_exception(e, request, attempt, max_attempts)
                if attempt == max_attempts - 1:
                    raise
                continue

        # This should not be reached, but just in case
        raise RequestException(
            f"Failed to complete request after {max_attempts} attempts"
        )

    def _handle_github_rate_limit(
        self, response: requests.Response, attempt: int, max_attempts: int
    ) -> None:
        """Handle GitHub API rate limiting with appropriate backoff."""
        if attempt >= max_attempts - 1:
            logger.error(
                "GitHub API rate limit exceeded after %d attempts for %s",
                max_attempts,
                response.request.url or "<unknown>",
            )
            return

        # Check for reset time in headers
        reset_time = response.headers.get("X-RateLimit-Reset")
        if reset_time:
            try:
                reset_timestamp = int(reset_time)
                current_time = int(time.time())
                wait_time = min(
                    reset_timestamp - current_time + 5, 300
                )  # Max 5 minutes
                if wait_time > 0:
                    logger.warning(
                        "GitHub API rate limit hit for %s. Waiting %d seconds until reset.",
                        response.request.url or "<unknown>",
                        wait_time,
                    )
                    time.sleep(wait_time)
                    return
            except (ValueError, TypeError):
                logger.debug("Could not parse GitHub rate limit reset time")

        # Fallback to exponential backoff
        wait_time = min(2**attempt + random.uniform(0, 1), 60)
        logger.warning(
            "GitHub API rate limit hit for %s. Retrying in %.1f seconds (attempt %d/%d)",
            response.request.url or "<unknown>",
            wait_time,
            attempt + 1,
            max_attempts,
        )
        time.sleep(wait_time)

    def _handle_retryable_exception(
        self,
        exception: Exception,
        request: requests.PreparedRequest,
        attempt: int,
        max_attempts: int,
    ) -> None:
        """Handle retryable exceptions with exponential backoff."""
        if attempt >= max_attempts - 1:
            logger.error(
                "Request failed after %d attempts for %s: %s",
                max_attempts,
                request.url or "<unknown>",
                exception,
            )
            return

        # Calculate backoff time with jitter
        wait_time = min(2**attempt + random.uniform(0, 1), 60)

        logger.warning(
            "Request failed for %s: %s. Retrying in %.1f seconds (attempt %d/%d)",
            request.url or "<unknown>",
            exception,
            wait_time,
            attempt + 1,
            max_attempts,
        )
        time.sleep(wait_time)


def create_retry_session(
    retry_config: dict[str, typing.Any] | None = None,
    timeout: float = 60.0,
) -> requests.Session:
    """Create a requests Session with retry capabilities.

    Args:
        retry_config: Configuration for retry behavior. If None, uses DEFAULT_RETRY_CONFIG.
        timeout: Default timeout for requests in seconds.

    Returns:
        A configured requests.Session with retry capabilities.
    """
    session = requests.Session()
    adapter = RetryHTTPAdapter(retry_config=retry_config, timeout=timeout)

    session.mount("http://", adapter)
    session.mount("https://", adapter)

    return session


def retry_on_exception(
    exceptions: tuple[type[Exception], ...] = RETRYABLE_EXCEPTIONS,
    max_attempts: int = 5,
    backoff_factor: float = 1.0,
    max_backoff: float = 60.0,
) -> typing.Callable[
    [typing.Callable[..., typing.Any]], typing.Callable[..., typing.Any]
]:
    """Decorator to retry a function on specific exceptions.

    Args:
        exceptions: Tuple of exception types that should trigger a retry.
        max_attempts: Maximum number of attempts.
        backoff_factor: Factor for exponential backoff.
        max_backoff: Maximum backoff time in seconds.

    Returns:
        Decorator function.
    """

    def decorator(
        func: typing.Callable[..., typing.Any],
    ) -> typing.Callable[..., typing.Any]:
        def wrapper(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
            for attempt in range(max_attempts):
                try:
                    return func(*args, **kwargs)
                except exceptions as e:
                    if attempt >= max_attempts - 1:
                        logger.error(
                            "Function %s failed after %d attempts: %s",
                            func.__name__,
                            max_attempts,
                            e,
                        )
                        raise

                    wait_time = min(
                        backoff_factor * (2**attempt) + random.uniform(0, 1),
                        max_backoff,
                    )

                    logger.warning(
                        "Function %s failed: %s. Retrying in %.1f seconds (attempt %d/%d)",
                        func.__name__,
                        e,
                        wait_time,
                        attempt + 1,
                        max_attempts,
                    )
                    time.sleep(wait_time)

            # This should not be reached due to the raise in the exception handler
            raise RuntimeError(f"Retry logic failed for {func.__name__}")

        return wrapper

    return decorator


# For backward compatibility and ease of use
def get_retry_session() -> requests.Session:
    """Get a pre-configured retry session with sensible defaults.

    This is the recommended way to get a session for general use.
    """
    return create_retry_session()



from __future__ import annotations

import collections
import logging
import os
import pathlib
import shutil
import sys
import tempfile
import textwrap
import typing
import zipfile
from urllib.parse import unquote, urlparse

import elfdeps
import tomlkit
import wheel.wheelfile  # type: ignore
from packaging.requirements import Requirement
from packaging.tags import Tag
from packaging.utils import BuildTag, canonicalize_name, parse_wheel_filename
from packaging.version import Version

from . import (
    external_commands,
    metrics,
    overrides,
    requirements_file,
    resolver,
    sources,
)

if typing.TYPE_CHECKING:
    from . import build_environment, context

logger = logging.getLogger(__name__)

FROMAGER_BUILD_SETTINGS = "fromager-build-settings"
FROMAGER_ELF_PROVIDES = "fromager-elf-provides.txt"
FROMAGER_ELF_REQUIRES = "fromager-elf-requires.txt"
FROMAGER_BUILD_REQ_PREFIX = "fromager"


def _extra_metadata_elfdeps(
    ctx: context.WorkContext,
    req: Requirement,
    wheel_root_dir: pathlib.Path,
    dist_info_dir: pathlib.Path,
) -> typing.Iterable[elfdeps.ELFInfo]:
    """Analyze a wheel's ELF dependencies and add info files

    Logs and returns library dependencies and library provides. Writes
    requirements to dist-info.
    """
    # mapping of required libraries to list of versions
    requires: set[elfdeps.SOInfo] = set()
    provides: set[elfdeps.SOInfo] = set()
    elfinfos: list[elfdeps.ELFInfo] = []

    settings = elfdeps.ELFAnalyzeSettings(filter_soname=True)
    for info in elfdeps.analyze_dirtree(wheel_root_dir, settings=settings):
        if info.filename is not None:
            relname = str(info.filename.relative_to(wheel_root_dir))
        else:
            relname = "n/a"
        logger.debug(
            f"{relname} ({info.soname}) "
            f"requires {sorted(info.requires)}, "
            f"provides {sorted(info.provides)}"
        )
        provides.update(info.provides)
        requires.update(info.requires)
        elfinfos.append(info)

    # Don't list provided names as requirements
    requires = requires.difference(provides)

    if requires:
        reqmap: dict[str, list[str]] = collections.defaultdict(list)
        for r in requires:
            reqmap[r.soname].append(r.version)

        names = sorted(
            name for name in reqmap if not name.startswith(("ld-linux", "rtld"))
        )
        logger.info("Requires libraries: %s", ", ".join(names))
        for name, versions in sorted(reqmap.items()):
            logger.debug(
                "Requires %s(%s)",
                name,
                ", ".join(v for v in versions if v),
            )

        requires_file = dist_info_dir / FROMAGER_ELF_REQUIRES
        with requires_file.open("w", encoding="utf-8") as f:
            for soinfo in sorted(requires):
                f.write(f"{soinfo}\n")

    if provides:
        names = sorted(p.soname for p in provides)
        logger.info("Provides libraries: %s", ", ".join(names))

        provides_file = dist_info_dir / FROMAGER_ELF_PROVIDES
        with provides_file.open("w", encoding="utf-8") as f:
            for soinfo in sorted(provides):
                f.write(f"{soinfo}\n")

    return elfinfos


def extract_info_from_wheel_file(
    req: Requirement, wheel_file: pathlib.Path
) -> tuple[str, Version, BuildTag, frozenset[Tag]]:
    # parse_wheel_filename normalizes the dist name, however the dist-info
    # directory uses the verbatim distribution name from the wheel file.
    # Packages with upper case names like "MarkupSafe" are affected.
    dist_name_normalized, dist_version, build_tag, wheel_tags = parse_wheel_filename(
        wheel_file.name
    )
    dist_name = wheel_file.name.split("-", 1)[0]
    if dist_name_normalized != canonicalize_name(dist_name):
        # sanity check, should never fail
        raise ValueError(f"{dist_name_normalized} does not match {dist_name}")
    return (dist_name, dist_version, build_tag, wheel_tags)


def default_add_extra_metadata_to_wheels(
    ctx: context.WorkContext,
    req: Requirement,
    version: Version,
    extra_environ: dict[str, str],
    sdist_root_dir: pathlib.Path,
    dist_info_dir: pathlib.Path,
) -> dict[str, typing.Any]:
    raise NotImplementedError


@metrics.timeit(description="add extra metadata to wheels")
def add_extra_metadata_to_wheels(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    version: Version,
    extra_environ: dict[str, str],
    sdist_root_dir: pathlib.Path,
    wheel_file: pathlib.Path,
) -> pathlib.Path:
    pbi = ctx.package_build_info(req)
    dist_name, dist_version, _, wheel_tags = extract_info_from_wheel_file(
        req, wheel_file
    )
    dist_filename = f"{dist_name}-{dist_version}"

    extra_data_plugin = overrides.find_override_method(
        req.name, "add_extra_metadata_to_wheels"
    )
    data_to_add = {}

    with tempfile.TemporaryDirectory() as dir_name:
        wheel_root_dir = pathlib.Path(dir_name) / dist_filename
        wheel_root_dir.mkdir()
        with zipfile.ZipFile(str(wheel_file)) as zf:
            for infolist in zf.filelist:
                # Check for path traversal attempts
                if (
                    os.path.isabs(infolist.filename)
                    or ".." in pathlib.Path(infolist.filename).parts
                ):
                    raise ValueError(f"Unsafe path in wheel: {infolist.filename}")
                zf.extract(infolist, wheel_root_dir)
                # the higher 16 bits store the permissions and type of file (i.e. stat.filemode)
                # the lower bits of this give us the permission
                permissions = infolist.external_attr >> 16 & 0o777
                wheel_root_dir.joinpath(infolist.filename).chmod(permissions)

        dist_info_dir = wheel_root_dir / f"{dist_filename}.dist-info"
        if not dist_info_dir.is_dir():
            raise ValueError(f"{wheel_file} does not contain {dist_info_dir.name}")

        if extra_data_plugin:
            data_to_add = overrides.invoke(
                extra_data_plugin,
                ctx=ctx,
                req=req,
                version=version,
                extra_environ=extra_environ,
                sdist_root_dir=sdist_root_dir,
                dist_info_dir=dist_info_dir,
            )
            if not isinstance(data_to_add, dict):
                logger.warning(
                    "unexpected return type from plugin add_extra_metadata_to_wheels. Expected dictionary. Will ignore"
                )
                data_to_add = {}

        if pbi.has_config:
            settings = pbi.serialize(mode="json", exclude_defaults=False)
        else:
            settings = {}
        if data_to_add:
            settings["metadata-from-plugin"] = data_to_add
        build_file = dist_info_dir / FROMAGER_BUILD_SETTINGS
        build_file.write_text(tomlkit.dumps(settings))

        req_files = sdist_root_dir.parent.glob("*-requirements.txt")
        for req_file in req_files:
            shutil.copy(
                req_file, dist_info_dir / f"{FROMAGER_BUILD_REQ_PREFIX}-{req_file.name}"
            )

        if any(tag.platform != "all" for tag in wheel_tags):
            # platlib wheel
            if sys.platform == "linux":
                _extra_metadata_elfdeps(
                    ctx=ctx,
                    req=req,
                    wheel_root_dir=wheel_root_dir,
                    dist_info_dir=dist_info_dir,
                )
            else:
                logger.debug(
                    "shared library dependency analysis not implemented for %s",
                    sys.platform,
                )
        else:
            logger.debug("%s is a purelib wheel", req.name)

        build_tag_from_settings = pbi.build_tag(version)
        build_tag = build_tag_from_settings if build_tag_from_settings else (0, "")

        cmd = [
            "wheel",
            "pack",
            str(wheel_root_dir),
            "--dest-dir",
            str(wheel_file.parent),
            "--build-number",
            f"{build_tag[0]}{build_tag[1]}",
        ]

        external_commands.run(
            cmd,
            cwd=dir_name,
            network_isolation=ctx.network_isolation,
        )

    wheel_file.unlink(missing_ok=True)
    wheels = list(wheel_file.parent.glob(f"{dist_filename}-*.whl"))
    if wheels:
        logger.info(
            f"added extra metadata and build tag {build_tag}, wheel renamed from {wheel_file.name} to {wheels[0].name}"
        )
        return wheels[0]
    raise FileNotFoundError("Could not locate new wheels file")


@metrics.timeit(description="build wheels")
def build_wheel(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    sdist_root_dir: pathlib.Path,
    version: Version,
    build_env: build_environment.BuildEnvironment,
) -> pathlib.Path:
    pbi = ctx.package_build_info(req)
    logger.info(
        f"building {ctx.variant} wheel for {req} in {sdist_root_dir} "
        f"writing to {ctx.wheels_build}"
    )

    # add package and variant env vars, package's parallel job vars, and
    # build_env's virtual env vars.
    extra_environ = pbi.get_extra_environ(build_env=build_env)

    if (
        pbi.build_ext_parallel
        and "DIST_EXTRA_CONFIG" not in extra_environ
        and "MAX_JOBS" in extra_environ
    ):
        # configure setuptools to use parallel builds
        # https://setuptools.pypa.io/en/latest/deprecated/distutils/configfile.html
        dist_extra_cfg = build_env.path / "dist-extra.cfg"
        dist_extra_cfg.write_text(
            textwrap.dedent(
                f"""
                [build_ext]
                parallel = {extra_environ["MAX_JOBS"]}
                """
            )
        )
        extra_environ["DIST_EXTRA_CONFIG"] = str(dist_extra_cfg)

    overrides.find_and_invoke(
        req.name,
        "build_wheel",
        default_build_wheel,
        ctx=ctx,
        build_env=build_env,
        extra_environ=extra_environ,
        req=req,
        sdist_root_dir=sdist_root_dir,
        build_dir=pbi.build_dir(sdist_root_dir),
        version=version,
    )
    # End the timer
    wheels = list(ctx.wheels_build.glob("*.whl"))
    if len(wheels) != 1:
        raise FileNotFoundError(
            f"Expected 1 built wheel in {ctx.wheels_build}, got {len(wheels)}"
        )

    wheel = add_extra_metadata_to_wheels(
        ctx=ctx,
        req=req,
        version=version,
        extra_environ=extra_environ,
        sdist_root_dir=sdist_root_dir,
        wheel_file=wheels[0],
    )
    return wheel


def default_build_wheel(
    ctx: context.WorkContext,
    build_env: build_environment.BuildEnvironment,
    extra_environ: dict[str, str],
    req: Requirement,
    sdist_root_dir: pathlib.Path,
    version: Version,
    build_dir: pathlib.Path,
) -> None:
    logger.debug(f"building wheel in {build_dir} with {extra_environ}")
    pbi = ctx.package_build_info(req)

    cmd = [
        os.fspath(build_env.python),
        "-m",
        "pip",
        "-vvv",
        "--disable-pip-version-check",
        "wheel",
        "--no-build-isolation",
        "--only-binary",
        ":all:",
        "--wheel-dir",
        os.fspath(ctx.wheels_build),
        "--no-deps",
        "--index-url",
        ctx.wheel_server_url,  # probably redundant, but just in case
        "--log",
        os.fspath(build_dir.parent / "build.log"),
    ]
    # config settings needs pip >= 24.0 to work. Fromager uses `virtualenv``
    # package to create virtual envs, which comes with recent pip. Stdlib's
    # `venv` comes with rather old pip.
    for key, values in pbi.config_settings.items():
        for value in values:
            cmd.append(f"--config-settings={key}={value}")
    cmd.append(os.fspath(build_dir))

    with tempfile.TemporaryDirectory() as dir_name:
        build_env.run(
            cmd,
            cwd=dir_name,
            extra_environ=extra_environ,
            network_isolation=ctx.network_isolation,
        )


def download_wheel(
    req: Requirement,
    wheel_url: str,
    output_directory: pathlib.Path,
) -> pathlib.Path:
    wheel_filename = output_directory / unquote(
        os.path.basename(urlparse(wheel_url).path)
    )
    if not wheel_filename.exists():
        logger.info(f"downloading pre-built wheel {wheel_url}")
        wheel_filename = _download_wheel_check(req, output_directory, wheel_url)
        logger.info(f"saved wheel to {wheel_filename}")
    else:
        logger.info(f"have existing wheel {wheel_filename}")

    return wheel_filename


def _download_wheel_check(
    req: Requirement, destination_dir: pathlib.Path, wheel_url: str
) -> pathlib.Path:
    wheel_filename = sources.download_url(
        req=req,
        destination_dir=destination_dir,
        url=wheel_url,
    )
    # validates whether the wheel is correct or not. will raise an error in the wheel is invalid
    wheel.wheelfile.WheelFile(wheel_filename)
    return wheel_filename


def get_wheel_server_urls(ctx: context.WorkContext, req: Requirement) -> list[str]:
    pbi = ctx.package_build_info(req)
    if pbi.wheel_server_url:
        # use only the wheel server from settings if it is defined. Do not fallback to other URLs
        servers = [pbi.wheel_server_url]
    else:
        servers = [resolver.PYPI_SERVER_URL]
        if ctx.wheel_server_url:
            servers.insert(0, ctx.wheel_server_url)
    return servers


@metrics.timeit(description="resolve wheel")
def resolve_prebuilt_wheel(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    wheel_server_urls: list[str],
    req_type: requirements_file.RequirementType | None = None,
) -> tuple[str, Version]:
    "Return URL to wheel and its version."
    for url in wheel_server_urls:
        try:
            wheel_url, resolved_version = resolver.resolve(
                ctx=ctx,
                req=req,
                sdist_server_url=url,
                include_sdists=False,
                include_wheels=True,
                req_type=req_type,
                # pre-built wheels must match platform
                ignore_platform=False,
            )
        except Exception:
            continue
        if wheel_url and resolved_version:
            return (wheel_url, resolved_version)
    raise ValueError(
        f"Could not find a prebuilt wheel for {req} on {' or '.join(wheel_server_urls)}"
    )



import pytest
from packaging.requirements import Requirement
from packaging.utils import canonicalize_name
from packaging.version import Version

from fromager import dependency_graph, requirements_file

raw_graph = {
    "": {
        "download_url": "",
        "pre_built": False,
        "version": "0",
        "canonicalized_name": "",
        "edges": [{"key": "a==2.0", "req_type": "install", "req": "a==2.0"}],
    },
    "a==2.0": {
        "download_url": "url",
        "pre_built": False,
        "version": "2.0",
        "canonicalized_name": "a",
        "edges": [
            {"key": "b==3.0", "req_type": "build-system", "req": "b==3.0"},
            {"key": "c==4.0", "req_type": "build-backend", "req": "c==4.0"},
        ],
    },
    "b==3.0": {
        "download_url": "url for b",
        "pre_built": False,
        "version": "3.0",
        "canonicalized_name": "b",
        "edges": [
            {"key": "c==4.0", "req_type": "build-sdist", "req": "c<=4.0"},
        ],
    },
    "c==4.0": {
        "download_url": "url for c",
        "pre_built": False,
        "version": "4.0",
        "canonicalized_name": "c",
        "edges": [],
    },
}


def test_graph_add_dependency():
    graph = dependency_graph.DependencyGraph()
    # top level dependency
    graph.add_dependency(
        parent_name=None,
        parent_version=None,
        req_type=requirements_file.RequirementType.INSTALL,
        req=Requirement("a==2.0"),
        req_version=Version("2.0"),
        download_url="url",
    )

    # children for toplevel
    graph.add_dependency(
        parent_name=canonicalize_name("a"),
        parent_version=Version("2.0"),
        req_type=requirements_file.RequirementType.BUILD_SYSTEM,
        req=Requirement("b==3.0"),
        req_version=Version("3.0"),
        download_url="url for b",
    )

    graph.add_dependency(
        parent_name=canonicalize_name("a"),
        parent_version=Version("2.0"),
        req_type=requirements_file.RequirementType.BUILD_BACKEND,
        req=Requirement("c==4.0"),
        req_version=Version("4.0"),
        download_url="url for c",
    )

    graph.add_dependency(
        parent_name=canonicalize_name("b"),
        parent_version=Version("3.0"),
        req_type=requirements_file.RequirementType.BUILD_SDIST,
        req=Requirement("c<=4.0"),
        req_version=Version("4.0"),
        download_url="url for c",
    )

    with pytest.raises(ValueError):
        # add dependency for a parent that doesn't exist
        graph.add_dependency(
            parent_name=canonicalize_name("z"),
            parent_version=Version("3.0"),
            req_type=requirements_file.RequirementType.BUILD_SYSTEM,
            req=Requirement("b==3.0"),
            req_version=Version("3.0"),
            download_url="url for b",
        )

    assert graph._to_dict() == raw_graph


def test_graph_from_dict():
    graph = dependency_graph.DependencyGraph.from_dict(raw_graph)
    assert graph._to_dict() == raw_graph


def test_get_install_dependencies():
    graph = dependency_graph.DependencyGraph.from_dict(raw_graph)
    graph.add_dependency(
        parent_name=canonicalize_name("a"),
        parent_version=Version("2.0"),
        req_type=requirements_file.RequirementType.INSTALL,
        req=Requirement("d>=4.0"),
        req_version=Version("6.0"),
        download_url="url for d",
    )

    # shouldn't be picked up by get_install_dependencies since b only appears as a build req
    graph.add_dependency(
        parent_name=canonicalize_name("b"),
        parent_version=Version("3.0"),
        req_type=requirements_file.RequirementType.INSTALL,
        req=Requirement("e>=4.0"),
        req_version=Version("6.0"),
        download_url="url for e",
    )

    install_nodes = [
        f"{node.to_dict()['canonicalized_name']}=={node.to_dict()['version']}"
        for node in graph.get_install_dependencies()
    ]
    assert install_nodes == ["a==2.0", "d==6.0"]

    # make b appear as install dependency
    graph.add_dependency(
        parent_name=None,
        parent_version=None,
        req_type=requirements_file.RequirementType.INSTALL,
        req=Requirement("b<4.0"),
        req_version=Version("3.0"),
        download_url="url for b",
    )

    install_nodes = [
        f"{node.to_dict()['canonicalized_name']}=={node.to_dict()['version']}"
        for node in graph.get_install_dependencies()
    ]
    assert install_nodes == ["a==2.0", "d==6.0", "b==3.0", "e==6.0"]


def test_cycles_get_install_dependencies():
    graph = dependency_graph.DependencyGraph.from_dict(raw_graph)
    # create cycle: a depends on d and d depends on a
    graph.add_dependency(
        parent_name=canonicalize_name("a"),
        parent_version=Version("2.0"),
        req_type=requirements_file.RequirementType.INSTALL,
        req=Requirement("d>=4.0"),
        req_version=Version("6.0"),
        download_url="url for d",
    )

    graph.add_dependency(
        parent_name=canonicalize_name("d"),
        parent_version=Version("6.0"),
        req_type=requirements_file.RequirementType.INSTALL,
        req=Requirement("a<=2.0"),
        req_version=Version("2.0"),
        download_url="url for a",
    )

    # add another duplicate toplevel
    graph.add_dependency(
        parent_name=None,
        parent_version=None,
        req_type=requirements_file.RequirementType.TOP_LEVEL,
        req=Requirement("a<=2.0"),
        req_version=Version("2.0"),
        download_url="url for a",
    )

    install_nodes = [
        f"{node.to_dict()['canonicalized_name']}=={node.to_dict()['version']}"
        for node in graph.get_install_dependencies()
    ]
    assert install_nodes == ["a==2.0", "d==6.0"]
